"""
Manage the Vault (or OpenBao) PKI secret engine and Vault-issued X.509 certificates.

.. versionadded:: 1.1.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import base64
import logging
import os

import salt.utils.files
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils.vault.helpers import filter_state_internal_kwargs
from saltext.vault.utils.vault.helpers import timestring_map
from saltext.vault.utils.vault.pki import check_cert_for_changes

try:
    import salt.utils.x509 as x509util

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False


log = logging.getLogger(__name__)

__virtualname__ = "vault_pki"


def __virtual__():
    if "x509.encode_certificate" not in __salt__:
        return (
            False,
            "x509_v2 needs to be explicitly enabled by setting `x509_v2: true` "
            "in the minion configuration value `features` until Salt 3008 (Argon).",
        )
    if not HAS_CRYPTOGRAPHY:
        return (False, "Could not load cryptography")
    return __virtualname__


VALID_FILE_ARGS = (
    "user",
    "group",
    "mode",
    "attrs",
    "makedirs",
    "dir_mode",
    "backup",
    "create",
    "follow_symlinks",
    "check_cmd",
    "tmp_dir",
    "tmp_ext",
    "selinux",
    "file_encoding",
    "encoding_errors",
    "win_owner",
    "win_perms",
    "win_deny_perms",
    "win_inheritance",
    "win_perms_reset",
)


def certificate_managed(
    name,
    common_name,
    role_name,
    private_key,
    mount="pki",
    ttl="720h",
    ttl_remaining="168h",
    issuer_ref=None,
    encoding="pem",
    append_ca_chain=False,
    sign_verbatim=False,
    private_key_passphrase=None,
    reissue=False,
    **kwargs,
):
    """
    Ensure an X.509 certificate is present as specified.

    .. note::
        The state can use ``sign-verbatim`` endpoint of Vault in which case CSR subject is fully
        translated. If not used, anything from CSR subject, except CN is ignored.
        Check `this issue <https://github.com/hashicorp/vault/issues/20719>`__ for more information.

    Required policy:

    .. code-block:: vaultpolicy

            # Need to read the role configuration in case of missing issuer_ref
            path "{mount}/roles/*" {
                capabilities = ["read"]
            }

            path "{mount}/issuer/{issuer_ref}/sign/{role_name}" {
                capabilities = ["update"]
            }
            # in case of sign_verbatim
            path "{mount}/issuer/{issuer_ref}/sign-verbatim/{role_name}" {
                capabilities = ["update"]
            }

    name
        Path to the certificate file.

    common_name
        Common name to be set for the certificate.

    role_name
        PKI role to be used for issuing the certificate from Vault.

    private_key
        Path or PEM formatted text of the private key used to sign CSR for the certificate.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    ttl
        Specifies the Time To Live value to be used for the validity period of the requested certificate,
        provided as a string duration with time suffix. Hour is the largest suffix. Defaults to ``720h`` or 30 days.

    ttl_remaining
        Specifies the Time To Live value to be used for checking remaining period before expiration
        after which certificate should be renewed.
        Provided as a string duration with time suffix. Hour is the largest suffix. Defaults to ``168h`` or 7 days.

    issuer_ref
        Override role's issuer for the certificate. Defaults to the one specified in the role.

    encoding
        Encoding to be used for the certificate file. Valid options are ``pem``, ``pkcs7_pem``, ``der``, ``pkcs7_der``. Defaults to ``pem``.

    append_ca_chain
        If set to true will append CA chain to the certificate. Defaults to ``false``.

        .. note::
            This will append all CA certificates except self-signed (as they shouldn't be in the chain anyway)!

    sign_verbatim
        If set to true, the resulting certificate follows the CSR exactly.
        Otherwise, only ``CN`` can be set for the subject, any other subject parameters (like ``O``) are ignored.

        .. warning::
            This option is using a potentially dangerous endpoint. Be careful when using that option, as roles
            are not restricting what can be issued anymore.

    private_key_passphrase
        Password for the private key if encrypted.

    reissue
        Always reissue the certificate. Defaults to ``false``.

    kwargs
        Any other parameter accepted by ``file_managed`` execution module or Vault PKI
        :obj:`sign_certificate <saltext.vault.modules.vault_pki.sign_certificate>` execution module.
    """

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The certificate is in the correct state",
    }

    changes = {}
    ca_chain = []
    verb = "create"
    file_args, cert_args = _split_file_kwargs(filter_state_internal_kwargs(kwargs))

    try:
        if encoding not in ["der", "pem", "pkcs7_der", "pkcs7_pem"]:
            raise SaltInvocationError(
                f"Invalid value '{encoding}' for encoding. Valid: der, pem, pkcs7_der, pkcs7_pem"
            )

        if timestring_map(ttl_remaining, cast=int) >= timestring_map(ttl, cast=int):
            raise SaltInvocationError("The ttl_remaning cannot be larger or equal to ttl.")

        # check file.managed changes early to avoid using unnecessary resources
        file_managed_test = _file_managed(name, test=True, replace=False, **file_args)
        if file_managed_test["result"] is False:
            ret["result"] = False
            ret["comment"] = "Problem while testing file.managed changes, see its output"
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if "is not present and is not set for creation" in file_managed_test["comment"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        # handle follow_symlinks
        if __salt__["file.is_link"](name):
            if file_args.get("follow_symlinks", True):
                name = os.path.realpath(name)
            else:
                # workaround https://github.com/saltstack/salt/issues/31802
                __salt__["file.remove"](name)
                changes["replaced"] = True

        replace = False
        file_exists = __salt__["file.file_exists"](name)

        if issuer_ref is None:
            issuer_ref = __salt__["vault_pki.read_role"](role_name, mount=mount)["issuer_ref"]
            if issuer_ref is None:
                raise CommandExecutionError(f"role {role_name} does not exists.")

        issuer_info = __salt__["vault_pki.read_issuer"](issuer_ref, mount=mount)

        if append_ca_chain:
            ca_chain = [x509util.load_cert(x) for x in issuer_info["ca_chain"]]
            # Filter self-signed CA, which shouldn't be in the chain.
            ca_chain = [
                cert
                for cert in ca_chain
                if cert.subject.rfc4514_string() != cert.issuer.rfc4514_string()
            ]

        if file_exists:
            if reissue:
                # No need to make any checks, just replace the cert
                changes["replaced"] = True
            else:

                changes = check_cert_for_changes(
                    current=name,
                    append_chain=ca_chain,
                    common_name=common_name,
                    encoding=encoding,
                    issuer=issuer_info["certificate"],
                    private_key=private_key,
                    private_key_passphrase=private_key_passphrase,
                    common_name_only=not sign_verbatim,
                    expire_tolerance=ttl_remaining,
                    **cert_args,
                )

        else:
            changes["created"] = True

        if not changes and file_managed_test["result"] and not file_managed_test["changes"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        ret["changes"] = changes
        if changes and file_exists:
            verb = "reissue"

        if __opts__["test"]:
            ret["result"] = None if changes else True
            ret["comment"] = (
                f"The certificate would have been {verb}d" if changes else ret["comment"]
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if changes:
            if not set(changes) - {
                "ca_chain",
                "encoding",
            }:
                verb = "recreate"
                cert = __salt__["x509.encode_certificate"](
                    name,
                    append_certs=ca_chain,
                    encoding=encoding,
                )
            else:
                issued_cert = __salt__["vault_pki.sign_certificate"](
                    common_name=common_name,
                    role_name=role_name,
                    private_key=private_key,
                    private_key_passphrase=private_key_passphrase,
                    ttl=ttl,
                    issuer_ref=issuer_ref,
                    mount=mount,
                    sign_verbatim=sign_verbatim,
                    remove_roots_from_chain=False,
                    **cert_args,
                )
                cert = __salt__["x509.encode_certificate"](
                    issued_cert["certificate"],
                    append_certs=ca_chain if encoding in ["pem", "pkcs7_pem"] else [],
                    encoding=encoding,
                )

            ret["comment"] = f"The certificate has been {verb}d"

            if encoding not in ["pem", "pkcs7_pem"]:
                # file.managed does not support binary contents, so create
                # an empty file first (makedirs). This will not work with check_cmd!
                file_managed_ret = _file_managed(name, replace=False, **file_args)
                _add_sub_state_run(ret, file_managed_ret)
                if not _check_file_ret(file_managed_ret, ret, name):
                    return ret
                _safe_atomic_write(name, base64.b64decode(cert), file_args.get("backup", ""))

        if not changes or encoding in ["pem", "pkcs7_pem"]:
            replace = bool(encoding in ["pem", "pkcs7_pem"] and changes)
            contents = cert if replace else None
            file_managed_ret = _file_managed(name, contents=contents, replace=replace, **file_args)
            _add_sub_state_run(ret, file_managed_ret)
            if not _check_file_ret(file_managed_ret, ret, file_exists):
                return ret

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_managed(name, mount="pki", issuer_ref=None, ttl=None, max_ttl=None, **kwargs):
    """
    Ensures PKI role is present and configured as required.

    name
        The name of the role.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    issuer_ref
        Issuer reference for the role. Can be name, id or literal ``default``.

    ttl
        Specifies the Time To Live value to be used for the validity period of the requested certificate,
        provided as a string duration with time suffix. Hour is the largest suffix.
        The value specified is strictly used for future validity.
        If not set, uses the system default value or the value of ``max_ttl``, whichever is shorter.

    max_ttl
        Specifies the maximum Time To Live provided as a string duration with time suffix.
        Hour is the largest suffix. If not set, defaults to the system maximum lease TTL.

    kwargs
        Any other parameter accepted by Vault
        :obj:`write_role <saltext.vault.modules.vault_pki.write_role>` execution module or Vault update role API method.
    """

    ret = {
        "name": name,
        "result": True,
        "comment": "The role is present as specified",
        "changes": {},
    }

    kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    def _diff_params(current):
        nonlocal issuer_ref, ttl, max_ttl, kwargs
        diff_params = (
            ("issuer_ref", issuer_ref),
            ("ttl", timestring_map(ttl, cast=int)),
            ("max_ttl", timestring_map(max_ttl, cast=int)),
        )
        changed = {}
        for param, arg in diff_params:
            if arg is None:
                continue
            if current[param] != arg:
                changed.update(
                    {
                        param: {
                            "old": current.get(param),
                            "new": arg,
                        }
                    }
                )
        for param, arg in kwargs.items():
            if param in current and current[param] != arg:
                changed.update(
                    {
                        param: {
                            "old": current.get(param),
                            "new": arg,
                        }
                    }
                )
        return changed

    changes = {}

    try:
        current = __salt__["vault_pki.read_role"](name, mount=mount)

        if current:
            changes = _diff_params(current)
            if not changes:
                return ret

        ret["changes"].update(changes)
        if not current:
            ret["changes"]["created"] = name
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                f"PKI role `{name}` would have been {'updated' if current else 'created'}"
            )
            return ret
        __salt__["vault_pki.write_role"](
            name=name, mount=mount, issuer_ref=issuer_ref, ttl=ttl, max_ttl=max_ttl, **kwargs
        )

        ret["comment"] = f"PKI role `{name}` has been {'updated' if current else 'created'}"
    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_absent(name, mount="pki"):
    """
    Ensure PKI role is absent.

    name
        The name of the role.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    """

    ret = {
        "name": name,
        "result": True,
        "comment": "",
        "changes": {},
    }

    try:
        current = __salt__["vault_pki.read_role"](name, mount=mount)

        if current is None:
            ret["comment"] = f"Role `{name}` is already absent."
            return ret

        ret["changes"]["deleted"] = name

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"PKI role `{name}` would have been deleted"
            return ret

        __salt__["vault_pki.delete_role"](name, mount=mount)

        ret["comment"] = f"Connection `{name}` has been deleted."

    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def _split_file_kwargs(kwargs):
    file_args = {"show_changes": False}
    extra_args = {}
    for k, v in kwargs.items():
        if k in VALID_FILE_ARGS:
            file_args[k] = v
        else:
            extra_args[k] = v

    if "file_encoding" in file_args:
        file_args["encoding"] = file_args.pop("file_encoding")
    return file_args, extra_args


def _add_sub_state_run(ret, sub):
    sub["low"] = {
        "name": ret["name"],
        "state": "file",
        "__id__": __low__["__id__"],
        "fun": "managed",
    }
    if "sub_state_run" not in ret:
        ret["sub_state_run"] = []
    ret["sub_state_run"].append(sub)


def _file_managed(name, test=None, **kwargs):
    if test not in [None, True]:
        raise SaltInvocationError("test param can only be None or True")
    # work around https://github.com/saltstack/salt/issues/62590
    test = test or __opts__["test"]
    res = __salt__["state.single"]("file.managed", name, test=test, **kwargs)
    return res[next(iter(res))]


def _safe_atomic_write(dst, data, backup):
    """
    Create a temporary file with only user r/w perms and atomically
    copy it to the destination, honoring ``backup``.
    """
    tmp = salt.utils.files.mkstemp(prefix=salt.utils.files.TEMPFILE_PREFIX)
    with salt.utils.files.fopen(tmp, "wb") as tmp_:
        tmp_.write(data)
    salt.utils.files.copyfile(
        tmp, dst, __salt__["config.backup_mode"](backup), __opts__["cachedir"]
    )
    salt.utils.files.safe_rm(tmp)


def _check_file_ret(fret, ret, current):
    if fret["result"] is False:
        ret["result"] = False
        ret["comment"] = (
            f"Could not {'create' if not current else 'update'} file, see file.managed output"
        )
        ret["changes"] = {}
        return False
    return True



import os

import pytest
import salt.config


@pytest.fixture
def minion_opts(tmp_path):  # pragma: no cover
    """
    Default minion configuration with relative temporary paths to not
    require root permissions.
    """
    root_dir = tmp_path / "minion"
    opts = salt.config.DEFAULT_MINION_OPTS.copy()
    opts["__role"] = "minion"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/minion.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "minion")
    return opts


@pytest.fixture
def master_opts(tmp_path):  # pragma: no cover
    """
    Default master configuration with relative temporary paths to not
    require root permissions.
    """
    root_dir = tmp_path / "master"
    opts = salt.config.master_config(None)
    opts["__role"] = "master"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/master.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "master")
    return opts


@pytest.fixture
def syndic_opts(tmp_path):  # pragma: no cover
    """
    Default master configuration with relative temporary paths to not
    require root permissions.
    """
    root_dir = tmp_path / "syndic"
    opts = salt.config.DEFAULT_MINION_OPTS.copy()
    opts["syndic_master"] = "127.0.0.1"
    opts["__role"] = "minion"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/syndic.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "syndic")
    return opts



import pytest
from saltfactories.utils import random_string

from tests.support.mysql import MySQLImage
from tests.support.mysql import create_mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_container  # pylint: disable=unused-import
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_read
from tests.support.vault import vault_revoke
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "cache": {
                "backend": "disk",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture(scope="module")
def mysql_image():
    version = "10.5"
    return MySQLImage(
        name="mariadb",
        tag=version,
        container_id=random_string(f"mariadb-{version}-"),
    )


@pytest.fixture
def role_args_common():
    return {
        "db_name": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def testrole():
    return {
        "default_ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def teststaticrole(mysql_container):
    return {
        "db_name": "testdb",
        "rotation_period": 86400,
        "username": mysql_container.mysql_user,
    }


@pytest.fixture
def testreissuerole():
    return {
        "default_ttl": 180,
        "max_ttl": 180,
    }


@pytest.fixture
def testdb(mysql_container, container_host_ref):
    return {
        "plugin_name": "mysql-database-plugin",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": "testrole,teststaticrole,testreissuerole",
        "username": "root",
        "password": mysql_container.mysql_passwd,
    }


@pytest.fixture(scope="module", autouse=True)
def db_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("database")
    yield
    assert vault_disable_secret_engine("database")


@pytest.fixture
def connection_setup(testdb):
    try:
        vault_write("database/config/testdb", **testdb)
        assert "testdb" in vault_list("database/config")
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")


@pytest.fixture(params=[["testrole"]])
def roles_setup(connection_setup, request, role_args_common):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            role_args.update(role_args_common)
            vault_write(f"database/roles/{role_name}", **role_args)
            assert role_name in vault_list("database/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("database/roles"):
                vault_delete(f"database/roles/{role_name}")
                assert role_name not in vault_list("database/roles")


@pytest.fixture
def role_static_setup(connection_setup, teststaticrole):  # pylint: disable=unused-argument
    role_name = "teststaticrole"
    try:
        vault_write(f"database/static-roles/{role_name}", **teststaticrole)
        assert role_name in vault_list("database/static-roles")
        yield
    finally:
        if role_name in vault_list("database/static-roles"):
            vault_delete(f"database/static-roles/{role_name}")
            assert role_name not in vault_list("database/static-roles")


@pytest.fixture
def vault_db(states):
    try:
        yield states.vault_db
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")
        if "testrole" in vault_list("database/roles"):
            vault_delete("database/roles/testrole")
            assert "testrole" not in vault_list("database/roles")
        if "teststaticrole" in vault_list("database/static-roles"):
            vault_delete("database/static-roles/teststaticrole")
            assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.fixture
def connargs(mysql_container, container_host_ref):
    return {
        "plugin": "mysql",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": ["testrole", "teststaticrole", "testreissuerole"],
        "username": "root",
        "password": mysql_container.mysql_passwd,
        "rotate": False,
    }


@pytest.fixture
def roleargs():
    return {
        "connection": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def roleargs_static(mysql_container):
    return {
        "connection": "testdb",
        "username": mysql_container.mysql_user,
        "rotation_period": 86400,
    }


def test_connection_present(vault_db, connargs):
    ret = vault_db.connection_present("testdb", **connargs)
    assert ret.result
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "testdb"
    assert "testdb" in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_connection_present_no_changes(vault_db, connargs):
    ret = vault_db.connection_present("testdb", **connargs)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("connection_setup")
def test_connection_present_allowed_roles_change(vault_db, connargs):
    connargs["allowed_roles"] = ["testrole", "teststaticrole", "newrole"]
    ret = vault_db.connection_present("testdb", **connargs)
    assert ret.result
    assert ret.changes
    assert "allowed_roles" in ret.changes
    assert (
        vault_read("database/config/testdb")["data"]["allowed_roles"] == connargs["allowed_roles"]
    )


@pytest.mark.usefixtures("connection_setup")
def test_connection_present_new_param(vault_db, connargs):
    connargs["username_template"] = r"{{random 20}}"
    ret = vault_db.connection_present("testdb", **connargs)
    assert ret.result
    assert ret.changes
    assert "username_template" in ret.changes
    assert (
        vault_read("database/config/testdb")["data"]["connection_details"]["username_template"]
        == connargs["username_template"]
    )


def test_connection_present_test_mode(vault_db, connargs):
    ret = vault_db.connection_present("testdb", test=True, **connargs)
    assert ret.result is None
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "testdb"
    assert "testdb" not in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_connection_absent(vault_db):
    ret = vault_db.connection_absent("testdb")
    assert ret.result
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == "testdb"
    assert "testdb" not in vault_list("database/config")


def test_connection_absent_no_changes(vault_db):
    ret = vault_db.connection_absent("testdb")
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("connection_setup")
def test_connection_absent_test_mode(vault_db):
    ret = vault_db.connection_absent("testdb", test=True)
    assert ret.result is None
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == "testdb"
    assert "testdb" in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_role_present(vault_db, roleargs):
    ret = vault_db.role_present("testrole", **roleargs)
    assert ret.result
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "testrole"
    assert "testrole" in vault_list("database/roles")


@pytest.mark.usefixtures("roles_setup")
def test_role_present_no_changes(vault_db, roleargs):
    ret = vault_db.role_present("testrole", **roleargs)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("roles_setup")
def test_role_present_no_changes_with_time_string(vault_db, roleargs):
    roleargs["default_ttl"] = "1h"
    ret = vault_db.role_present("testrole", **roleargs)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("roles_setup")
def test_role_present_param_change(vault_db, roleargs):
    roleargs["default_ttl"] = 1337
    ret = vault_db.role_present("testrole", **roleargs)
    assert ret.result
    assert ret.changes
    assert "default_ttl" in ret.changes
    assert vault_read("database/roles/testrole")["data"]["default_ttl"] == 1337


@pytest.mark.usefixtures("connection_setup")
def test_role_present_test_mode(vault_db, roleargs):
    ret = vault_db.role_present("testrole", test=True, **roleargs)
    assert ret.result is None
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "testrole"
    assert "testrole" not in vault_list("database/roles")


@pytest.mark.usefixtures("connection_setup")
def test_static_role_present(vault_db, roleargs_static):
    ret = vault_db.static_role_present("teststaticrole", **roleargs_static)
    assert ret.result
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "teststaticrole"
    assert "teststaticrole" in vault_list("database/static-roles")


@pytest.mark.usefixtures("role_static_setup")
def test_static_role_present_no_changes(vault_db, roleargs_static):
    ret = vault_db.static_role_present("teststaticrole", **roleargs_static)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("role_static_setup")
def test_static_role_present_param_change(vault_db, roleargs_static):
    roleargs_static["rotation_period"] = 1337
    ret = vault_db.static_role_present("teststaticrole", **roleargs_static)
    assert ret.result
    assert ret.changes
    assert "rotation_period" in ret.changes
    assert vault_read("database/static-roles/teststaticrole")["data"]["rotation_period"] == 1337


@pytest.mark.usefixtures("connection_setup")
def test_static_role_present_test_mode(vault_db, roleargs_static):
    ret = vault_db.static_role_present("teststaticrole", test=True, **roleargs_static)
    assert ret.result is None
    assert ret.changes
    assert "created" in ret.changes
    assert ret.changes["created"] == "teststaticrole"
    assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.mark.usefixtures("roles_setup")
def test_role_absent(vault_db):
    ret = vault_db.role_absent("testrole")
    assert ret.result
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == "testrole"
    assert "testrole" not in vault_list("database/roles")


@pytest.mark.usefixtures("role_static_setup")
def test_role_absent_static(vault_db):
    ret = vault_db.role_absent("teststaticrole", static=True)
    assert ret.result
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == "teststaticrole"
    assert "teststaticrole" not in vault_list("database/static-roles")


def test_role_absent_no_changes(vault_db):
    ret = vault_db.role_absent("testrole")
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("roles_setup")
def test_role_absent_test_mode(vault_db):
    ret = vault_db.role_absent("testrole", test=True)
    assert ret.result is None
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"] == "testrole"
    assert "testrole" in vault_list("database/roles")


@pytest.fixture(params=(False, True))
def testmode(request):
    return request.param


@pytest.fixture(params=({},))
def _cached_creds(request, loaders, roles_setup):  # pylint: disable=unused-argument
    kwargs = {"name": "testrole", "cache": True}
    kwargs.update(request.param)
    ret = loaders.modules.vault_db.get_creds(**kwargs)
    assert ret
    assert "username" in ret
    assert "password" in ret
    assert loaders.modules.vault_db.list_cached()
    # We need to clear the context because in the test suite, the state modules
    # are running in a different one than the execution modules and the lease
    # has already been cached in the context of the execution module.
    # This means it does not pick up changes to the cached files, but we need
    # it to check changes in the tests.
    loaders.context.clear()
    return ret


@pytest.mark.usefixtures("roles_setup")
def test_creds_cached(testmode, vault_db, modules):
    ret = vault_db.creds_cached("testrole", test=testmode)
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "new" in ret.changes
    assert "issued" in ret.comment
    assert ("would have" in ret.comment) is testmode
    assert bool(modules.vault_db.list_cached()) is not testmode


def test_creds_cached_already_cached(testmode, vault_db, modules, _cached_creds):
    ret = vault_db.creds_cached("testrole", test=testmode)
    assert ret.result is True
    assert not ret.changes
    assert "already cached and valid" in ret.comment
    assert modules.vault_db.get_creds("testrole") == _cached_creds


def test_creds_cached_already_cached_but_different_params(
    testmode, vault_db, modules, _cached_creds
):
    """
    Ensure changed parameters are reported without reissuing the
    lease, if possible.
    """
    ret = vault_db.creds_cached(
        "testrole",
        valid_for=800,
        renew_increment=120,
        revoke_delay=240,
        meta={"foo": "bar"},
        test=testmode,
    )
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "new" not in ret.changes
    assert "expiry" not in ret.changes
    assert ret.changes == {
        "min_ttl": {"old": None, "new": 800},
        "renew_increment": {"old": None, "new": 120},
        "revoke_delay": {"old": None, "new": 240},
        "meta": {"old": None, "new": {"foo": "bar"}},
    }
    assert "edited" in ret.comment
    assert ("would have" in ret.comment) is testmode
    assert modules.vault_db.get_creds("testrole") == _cached_creds
    new = modules.vault_db.list_cached()
    new = new[next(iter(new))]
    assert (new["min_ttl"] == 800) is not testmode
    assert (new["renew_increment"] == 120) is not testmode
    assert (new["revoke_delay"] == 240) is not testmode
    assert (new["meta"] == {"foo": "bar"}) is not testmode


def test_creds_cached_renew(testmode, vault_db, modules, _cached_creds):
    """
    Ensure renewed credentials are reported as such.
    """
    ret = vault_db.creds_cached("testrole", valid_for="2h", test=testmode)
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "expiry" in ret.changes
    assert ret["changes"] == {"expiry": True, "min_ttl": {"old": None, "new": "2h"}}
    assert "renewed" in ret.comment
    assert ("would have" in ret.comment) is testmode
    if not testmode:
        assert modules.vault_db.get_creds("testrole") == _cached_creds
    # can't check the changes because requesting it would
    # apply the changes and listing the cache does not include the data


@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("roles_setup", [["testreissuerole"]], indirect=True)
@pytest.mark.parametrize(
    "_cached_creds", ({"name": "testreissuerole", "valid_for": 185},), indirect=True
)
def test_creds_cached_reissue(testmode, vault_db, modules, _cached_creds):
    """
    Ensure reissued credentials are reported as such.
    """
    ret = vault_db.creds_cached("testreissuerole", valid_for=160, test=testmode)
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "expiry" in ret.changes
    if testmode:
        # this is hard to detect in test mode TODO?
        assert "renewed/reissued" in ret.comment
        assert ret["changes"] == {"expiry": True, "min_ttl": {"old": 185, "new": 160}}
    else:
        assert "reissued" in ret.comment
        assert ret["changes"] == {
            "expiry": True,
            "min_ttl": {"old": 185, "new": 160},
            "reissued": True,
        }
    assert ("would have" in ret.comment) is testmode
    new = modules.vault_db.list_cached()
    new = new[next(iter(new))]
    assert (new["min_ttl"] == 160) is not testmode


@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("roles_setup", [["testreissuerole"]], indirect=True)
@pytest.mark.parametrize(
    "_cached_creds", ({"name": "testreissuerole", "valid_for": 185},), indirect=True
)
def test_creds_cached_reissue_only(testmode, vault_db, loaders, _cached_creds):
    """
    Ensure that expired leases are recognized, even if valid_for has not been set.
    """
    old = loaders.modules.vault_db.list_cached()
    old = old[next(iter(old))]
    old.pop("expires_in")
    loaders.context.clear()  # to get updated information because of differing ctx
    ret = vault_db.creds_cached("testreissuerole", test=testmode)
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "expiry" in ret.changes
    if testmode:
        # this is hard to detect in test mode TODO?
        assert "renewed/reissued" in ret.comment
        assert ret["changes"] == {"expiry": True}
    else:
        assert "reissued" in ret.comment
        assert ret["changes"] == {"expiry": True, "reissued": True}
    assert ("would have" in ret.comment) is testmode
    new = loaders.modules.vault_db.list_cached()
    new = new[next(iter(new))]
    new.pop("expires_in")
    assert (new == old) is testmode


def test_creds_uncached(testmode, vault_db, modules, _cached_creds):
    ret = vault_db.creds_uncached("testrole", test=testmode)
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert "revoked" in ret.changes
    assert ("would have" in ret.comment) is testmode
    assert "revoked" in ret.comment
    after = modules.vault_db.list_cached()
    assert bool(after) is testmode


def test_creds_uncached_no_changes(testmode, vault_db):
    ret = vault_db.creds_uncached("testrole", test=testmode)
    assert ret.result is True
    assert not ret.changes
    assert "No matching credentials" in ret.comment



"""
SSH wrapper for the :py:mod:`vault_db <saltext.vault.modules.vault_db>` execution module.

See there for documentation.
"""

from salt.utils.functools import namespaced_function

from saltext.vault.modules.vault_db import _write_role
from saltext.vault.modules.vault_db import clear_cached
from saltext.vault.modules.vault_db import delete_connection
from saltext.vault.modules.vault_db import delete_role
from saltext.vault.modules.vault_db import fetch_connection
from saltext.vault.modules.vault_db import fetch_role
from saltext.vault.modules.vault_db import get_creds
from saltext.vault.modules.vault_db import list_cached
from saltext.vault.modules.vault_db import list_connections
from saltext.vault.modules.vault_db import list_roles
from saltext.vault.modules.vault_db import renew_cached
from saltext.vault.modules.vault_db import reset_connection
from saltext.vault.modules.vault_db import rotate_root
from saltext.vault.modules.vault_db import rotate_static_role
from saltext.vault.modules.vault_db import write_connection
from saltext.vault.modules.vault_db import write_role
from saltext.vault.modules.vault_db import write_static_role

_write_role = namespaced_function(_write_role, globals())
clear_cached = namespaced_function(clear_cached, globals())
delete_connection = namespaced_function(delete_connection, globals())
delete_role = namespaced_function(delete_role, globals())
fetch_connection = namespaced_function(fetch_connection, globals())
fetch_role = namespaced_function(fetch_role, globals())
get_creds = namespaced_function(get_creds, globals())
list_cached = namespaced_function(list_cached, globals())
list_connections = namespaced_function(list_connections, globals())
list_roles = namespaced_function(list_roles, globals())
renew_cached = namespaced_function(renew_cached, globals())
reset_connection = namespaced_function(reset_connection, globals())
rotate_root = namespaced_function(rotate_root, globals())
rotate_static_role = namespaced_function(rotate_static_role, globals())
write_connection = namespaced_function(write_connection, globals())
write_role = namespaced_function(write_role, globals())
write_static_role = namespaced_function(write_static_role, globals())



import logging
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

import pytest

import saltext.vault.utils.vault as vaultutil
from saltext.vault.pillar import vault


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__utils__": {
                "vault.expand_pattern_lists": Mock(side_effect=lambda x, *args, **kwargs: [x])
            }
        }
    }


@pytest.fixture
def data():
    return {"foo": "bar"}


@pytest.fixture
def read_kv(data):
    with patch("saltext.vault.utils.vault.read_kv", autospec=True) as read:
        read.return_value = data
        yield read


@pytest.fixture
def read_kv_not_found(read_kv):
    read_kv.side_effect = vaultutil.VaultNotFoundError


@pytest.fixture
def role_a():
    return {
        "from_db": True,
        "pass": "hunter2",
        "list": ["a", "b"],
    }


@pytest.fixture
def role_b():
    return {
        "from_web": True,
        "pass": "hunter1",
        "list": ["c", "d"],
    }


def test_ext_pillar(read_kv, data):
    """
    Test ext_pillar functionality. KV v1/2 is handled by the utils module.
    """
    ext_pillar = vault.ext_pillar("testminion", {}, "secret/path")
    read_kv.assert_called_once_with("secret/path", opts=ANY, context=ANY)
    assert ext_pillar == data


@pytest.mark.usefixtures("read_kv_not_found")
def test_ext_pillar_not_found(caplog):
    """
    Test that HTTP 404 is handled correctly
    """
    with caplog.at_level(logging.INFO):
        ext_pillar = vault.ext_pillar("testminion", {}, "secret/path")
        assert ext_pillar == {}
        assert "Vault secret not found for: secret/path" in caplog.messages


@pytest.mark.usefixtures("read_kv")
def test_ext_pillar_nesting_key(data):
    """
    Test that nesting_key is honored as expected
    """
    ext_pillar = vault.ext_pillar("testminion", {}, "secret/path", nesting_key="baz")
    assert ext_pillar == {"baz": data}


@pytest.mark.parametrize(
    "pattern,expected",
    [
        ("no/template/in/use", ["no/template/in/use"]),
        ("salt/minions/{minion}", ["salt/minions/test-minion"]),
        ("salt/roles/{pillar[role]}", ["salt/roles/foo"]),
        ("salt/roles/{pillar[nonexistent]}", []),
    ],
)
def test_get_paths(pattern, expected):
    """
    Test that templated paths are resolved as expected.
    Expansion of lists is tested in the utility module unit test.
    """
    previous_pillar = {
        "role": "foo",
    }
    result = vault._get_paths(pattern, "test-minion", previous_pillar)
    assert result == expected


@pytest.mark.parametrize(
    "first,second,expected",
    [
        (
            "role_a",
            "role_b",
            {"from_db": True, "from_web": True, "list": ["c", "d"], "pass": "hunter1"},
        ),
        (
            "role_b",
            "role_a",
            {"from_db": True, "from_web": True, "list": ["a", "b"], "pass": "hunter2"},
        ),
    ],
)
def test_ext_pillar_merging(read_kv, first, second, expected, request):
    """
    Test that patterns that result in multiple paths are merged as expected.
    """
    first = request.getfixturevalue(first)
    second = request.getfixturevalue(second)
    read_kv.side_effect = (first, second)
    ext_pillar = vault.ext_pillar(
        "test-minion",
        {"roles": ["db", "web"]},
        path="salt/roles/{pillar[roles]}",
        merge_strategy="smart",
        merge_lists=False,
    )
    assert ext_pillar == expected


def test_ext_pillar_disabled_during_pillar_rendering(read_kv):
    """
    Ensure ext_pillar returns an empty dict when called during pillar
    template rendering to prevent a cyclic dependency.
    """
    extra = {"_vault_runner_is_compiling_pillar_templates": True}
    res = vault.ext_pillar("test-minion", {}, path="secret/path", extra_minion_data=extra)
    assert res == {}
    read_kv.assert_not_called()


@pytest.mark.parametrize("kwarg", (True, False))
@pytest.mark.parametrize(
    "conf",
    (
        "path=secret/path",
        "I have no idea why this was accepted before: path=secret/path",
    ),
)
def test_deprecated_config(read_kv, data, kwarg, conf):
    """
    Ensure the previous config with the ``path=`` prefix is still recognized,
    but warned about.
    """
    with pytest.deprecated_call(match="`path=`"):
        if kwarg:
            ext_pillar = vault.ext_pillar("testminion", {}, conf=conf)
        else:
            ext_pillar = vault.ext_pillar("testminion", {}, conf)
        read_kv.assert_called_once_with("secret/path", opts=ANY, context=ANY)
        assert ext_pillar == data



from unittest.mock import patch

import pytest
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

import saltext.vault.utils.vault as vaultutil
from saltext.vault.modules import vault_pki


@pytest.fixture
def configure_loader_modules():
    return {
        vault_pki: {
            "__grains__": {"id": "test-minion"},
        }
    }


@pytest.fixture
def data():
    return {"foo": "bar"}


@pytest.fixture
def data_role():
    return {
        "dummy": {
            "data": {
                "allow_any_name": True,
            }
        },
        "no-subddomains": {
            "data": {
                "allow_subdomains": False,
            }
        },
        "no-serverflag": {
            "data": {
                "server_flag": True,
            }
        },
    }


@pytest.fixture
def data_roles_list():
    return {"data": {"keys": ["foo"]}}


@pytest.fixture
def role_not_found(query):
    query.side_effect = vaultutil.VaultNotFoundError
    yield query


@pytest.fixture
def list_roles(data_roles_list):
    with patch("saltext.vault.utils.vault.query", autospec=True) as _list:
        _list.return_value = data_roles_list
        yield _list


@pytest.fixture
def read_role(data_role, role_name):
    with patch("saltext.vault.utils.vault.query", autospec=True) as _data:
        _data.return_value = data_role[role_name]
        yield _data


# @pytest.fixture
# def delete_role():
#     with patch("saltext.vault.utils.vault.query", autospec=True) as delete_role:
#         yield delete_role


@pytest.fixture
def query():
    with patch("saltext.vault.utils.vault.query", autospec=True) as _query:
        yield _query


@pytest.mark.usefixtures("list_roles")
@pytest.mark.parametrize(
    "expected",
    [["foo"]],
)
def test_list_roles(expected):
    res = vault_pki.list_roles()
    assert res == expected


@pytest.mark.usefixtures("read_role")
@pytest.mark.parametrize(
    "role_name, expected",
    [
        ("dummy", {"allow_any_name": True}),
        ("no-subddomains", {"allow_subdomains": False}),
        ("no-serverflag", {"server_flag": True}),
    ],
)
def test_read_role(role_name, expected):
    ret = vault_pki.read_role(role_name)
    assert ret == expected


@pytest.mark.parametrize("issuer", [None, "default", "someother"])
def test_write_role_payload(query, issuer):
    args = {
        "ttl": "300h",
        "max_ttl": "360h",
        "server_flag": True,
        "allow_subdomains": False,
        "allowed_domains": ["example.com", "saltproject.io"],
        "client_flag": True,
        "allow_localhost": True,
        "require_cn": False,
    }

    assert vault_pki.write_role("role", mount="mount", issuer_ref=issuer, **args) is True
    endpoint = query.call_args[0][1]
    payload = query.call_args[1]["payload"]
    assert endpoint == "mount/roles/role"
    expected_payload = args.copy()
    if issuer is not None:
        expected_payload["issuer_ref"] = issuer
    assert payload == expected_payload


def test_delete_role_payload(query):
    assert vault_pki.delete_role("role", mount="mount") is True
    endpoint = query.call_args[0][1]
    assert endpoint == "mount/roles/role"


@pytest.mark.usefixtures("role_not_found")
def test_write_role_raises_err():
    with pytest.raises(CommandExecutionError, match=".*VaultNotFoundError.*"):
        vault_pki.write_role("some/path")


@pytest.mark.usefixtures("role_not_found")
def test_delete_role_return_false_if_not_found():
    ret = vault_pki.delete_role("some/path")
    assert not ret


@pytest.mark.usefixtures("role_not_found")
def test_read_role_return_none_if_not_found():
    ret = vault_pki.read_role("some/path")
    assert ret is None


@pytest.mark.usefixtures("role_not_found")
def test_list_roles_return_empty_array_if_not_found():
    ret = vault_pki.list_roles("some/path")
    assert ret == []


@pytest.mark.parametrize(
    "common_name,root_type,args",
    [
        (
            "root_ca",
            "exported",
            {
                "issuer_name": "root-ca",
                "key-name": "root-ca-key",
                "max_path_length": 4,
                "key_bits": 384,
                "key_type": "ec",
            },
        ),
        (
            "root_ca",
            "internal",
            {
                "key_bits": 384,
                "key_type": "ec",
            },
        ),
        ("root_ca", "internal", {}),
    ],
)
def test_generate_root_payload(query, common_name, root_type, args):
    vault_pki.generate_root(common_name=common_name, type=root_type, mount="mount", **args)

    endpoint = query.call_args[0][1]
    payload = query.call_args[1]["payload"]
    assert endpoint == f"mount/root/generate/{root_type}"
    expected_payload = payload.copy()
    expected_payload["common_name"] = common_name
    assert payload == expected_payload


def test_generate_root_raise_err_with_default_name():
    with pytest.raises(SaltInvocationError):
        vault_pki.generate_root("my root", issuer_name="default")

    with pytest.raises(SaltInvocationError):
        vault_pki.generate_root("my root", key_name="default")


@pytest.mark.parametrize(
    "args", [{"serial": "00:11:22:33:44:55", "certificate": "-----BEGIN CERTIFICATE..."}, {}]
)
def test_revoke_certificate_raise_err(args):
    with pytest.raises(SaltInvocationError):
        vault_pki.revoke_certificate(**args)



import datetime

import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from salt.exceptions import SaltInvocationError

from saltext.vault.utils.vault import pki


class CAFixture:
    def __init__(self, common_name):
        self.common_name = common_name
        self.private_key = None
        self.certificate = None

    def generate(self):
        one_day = datetime.timedelta(1, 0, 0)
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        public_key = private_key.public_key()
        builder = x509.CertificateBuilder()
        builder = builder.subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, self.common_name),
                ]
            )
        )
        builder = builder.issuer_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, self.common_name),
                ]
            )
        )

        builder = builder.not_valid_before(datetime.datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.datetime.today() + (one_day * 30))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=1), critical=True
        )
        builder = builder.add_extension(
            x509.KeyUsage(
                crl_sign=True,
                digital_signature=True,
                key_cert_sign=True,
                content_commitment=False,
                data_encipherment=False,
                decipher_only=False,
                encipher_only=False,
                key_agreement=False,
                key_encipherment=False,
            ),
            critical=True,
        )
        builder = builder.add_extension(
            x509.SubjectAlternativeName([x509.DNSName("cryptography.io")]), critical=False
        )
        certificate = builder.sign(
            private_key=private_key,
            algorithm=hashes.SHA256(),
        )
        self.private_key = private_key
        self.certificate = certificate


class CertificateFixture:
    def __init__(self, common_name, ca: CAFixture):
        self.common_name = common_name
        self.ca = ca
        self.private_key = None
        self.certificate = None

    def generate(self, alt_names=None):
        one_day = datetime.timedelta(1, 0, 0)
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        public_key = private_key.public_key()
        builder = x509.CertificateBuilder()
        builder = builder.subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, self.common_name),
                ]
            )
        )
        builder = builder.issuer_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, self.common_name),
                ]
            )
        )

        builder = builder.not_valid_before(datetime.datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.datetime.today() + (one_day * 1))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)
        if alt_names is not None:
            builder = builder.add_extension(
                x509.SubjectAlternativeName([x509.DNSName(name) for name in alt_names]),
                critical=False,
            )
        builder = builder.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )

        certificate = builder.sign(
            private_key=self.ca.private_key,
            algorithm=hashes.SHA256(),
        )

        self.private_key = private_key
        self.certificate = certificate


class PKIInfra:
    certs = {}

    def __init__(self, name):
        ca = CAFixture(name)
        ca.generate()
        self.certs = {}
        self.ca = ca

    def issue_certificate(self, common_name, alt_names=None):
        if common_name not in self.certs:
            cert = CertificateFixture(f"{common_name}", self.ca)
            cert.generate(alt_names)
            self.certs[common_name] = cert
        return self.certs[common_name]


class PKIInfraFactory:
    pki_infra = {}

    @classmethod
    def instance(cls, name):
        if name not in cls.pki_infra:
            inst = PKIInfra(name=f"{name} CA")
            cls.pki_infra[name] = inst
        return cls.pki_infra[name]


@pytest.fixture(scope="module")
def existing_ca():
    instance = PKIInfraFactory.instance("existing")
    instance.issue_certificate("acme.com")

    return instance


@pytest.fixture(scope="module")
def existing_ca_with_alt_names():
    instance = PKIInfraFactory.instance("existing")
    instance.issue_certificate("acme.org", alt_names=["acme.com"])
    return instance


@pytest.fixture(scope="module")
def new_ca():
    instance = PKIInfraFactory.instance("new")
    instance.issue_certificate("acme.com")

    return instance


@pytest.fixture
def existing_pki(existing_ca):  # pylint: disable=unused-argument
    instance = PKIInfraFactory.instance("existing")
    cert_obj = instance.issue_certificate("acme.com")

    return cert_obj.certificate, cert_obj.private_key, [cert_obj.ca.certificate]


@pytest.fixture
def new_pki():
    instance = PKIInfraFactory.instance("new")
    cert_obj = instance.issue_certificate("acme.com")

    return cert_obj.certificate, cert_obj.private_key, [cert_obj.ca.certificate]


@pytest.mark.parametrize(
    "inpt,expected",
    [
        (60, "3C"),
        (4508375982735402, "10:04:58:14:F7:00:2A"),
        (123456789011, "1C:BE:99:1A:13"),
    ],
)
def test_dec2hex(inpt, expected):
    assert pki.dec2hex(inpt) == expected


@pytest.mark.parametrize(
    "inpt,match",
    [
        (-60, ".*non-negative"),
        ("00:11:22:33:44:55:66:77:88:99", ".*input must be integer.*"),
    ],
)
def test_dec2hex_raise_err(inpt, match):
    with pytest.raises(SaltInvocationError, match=match):
        pki.dec2hex(inpt)


def test_compare_ca_chain_with_new(existing_pki, new_pki):
    _, _, chain = existing_pki
    _, _, new_chain = new_pki
    assert pki.compare_ca_chain(chain, new_chain) is False


def test_compare_ca_chain_with_same(existing_pki):
    _, _, chain = existing_pki
    assert pki.compare_ca_chain(chain, chain) is True


def test_compare_ca_chain_with_same_diff_len(existing_pki):
    _, _, chain = existing_pki
    assert pki.compare_ca_chain(chain, chain + chain) is False



import os
import platform
import sys

DARKGREEN = (0, 100, 0)
DARKRED = (139, 0, 0)
YELLOW = (255, 255, 0)


def ensure_utf8():
    """
    On Windows, ensure stdout/stderr output uses UTF-8 encoding.
    """
    if platform.system() != "Windows":
        return
    for stream in (sys.stdout, sys.stderr):
        if stream.encoding != "utf-8":
            stream.reconfigure(encoding="utf-8")


def pprint(msg, bold=False, fg=None, bg=None, stream=None, force_print=False):
    """
    Ugly helper for printing a bit more fancy output.
    Stand-in for questionary/prompt_toolkit.
    """
    if not force_print and os.getenv("TOOLS_SILENT"):
        return
    out = ""
    if bold:
        out += "\033[1m"
    if fg:
        red, green, blue = fg
        out += f"\033[38;2;{red};{green};{blue}m"
    if bg:
        red, green, blue = bg
        out += f"\033[48;2;{red};{green};{blue}m"
    out += msg
    if bold or fg or bg:
        out += "\033[0m"
    print(out, file=stream or sys.stdout)


def status(msg, message=None):
    out = f"\n    → {msg}"
    pprint(out, bold=True, fg=DARKGREEN, stream=sys.stderr)
    if message:
        pprint(message, stream=sys.stderr)


def warn(header, message=None):
    out = f"\n{header}"
    pprint(out, bold=True, bg=DARKRED, stream=sys.stderr, force_print=True)
    if message:
        pprint(message, stream=sys.stderr, force_print=True)



"""
Polyfill for very basic ``plumbum`` functionality, no external libs required.
Makes scripts that call a lot of CLI commands much more pleasant to write.
"""

import os
import platform
import shlex
import shutil
import subprocess
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Union


class CommandNotFound(RuntimeError):
    """
    Raised when a command cannot be found in $PATH
    """


@dataclass(frozen=True)
class ProcessResult:
    """
    The full process result, returned by ``.run`` methods.
    The ``__call__`` ones just return stdout.
    """

    retcode: int
    stdout: str | bytes
    stderr: str | bytes
    argv: tuple

    def check(self, retcode=None):
        """
        Check if the retcode is expected. retcode can be a list.
        """
        if retcode is None:
            expected = [0]
        elif not isinstance(retcode, (list, tuple)):
            expected = [retcode]
        if self.retcode not in expected:
            raise ProcessExecutionError(self.argv, self.retcode, self.stdout, self.stderr)

    def __str__(self):
        msg = [
            "Process execution result:",
            f"Command: {shlex.join(self.argv)}",
            f"Retcode: {self.retcode}",
            "Stdout:   |",
        ]
        msg += [" " * 10 + "| " + line for line in str(self.stdout).splitlines()]
        msg.append("Stderr:   |")
        msg += [" " * 10 + "| " + line for line in str(self.stderr).splitlines()]
        return "\n".join(msg)


class ProcessExecutionError(OSError):
    """
    Raised by ProcessResult.check when an unexpected retcode was returned.
    """

    def __init__(self, argv, retcode, stdout, stderr):
        self.argv = argv
        self.retcode = retcode
        if isinstance(stdout, bytes):
            stdout = ascii(stdout)
        if isinstance(stderr, bytes):
            stderr = ascii(stderr)
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        msg = [
            "Process finished with unexpected exit code",
            f"Retcode: {self.retcode}",
            f"Command: {shlex.join(self.argv)}",
            "Stdout:   |",
        ]
        msg += [" " * 10 + "| " + line for line in str(self.stdout).splitlines()]
        msg.append("Stderr:   |")
        msg += [" " * 10 + "| " + line for line in str(self.stderr).splitlines()]
        return "\n".join(msg)


class Local:
    """
    Glue for command environment defaults.
    Should be treated as a singleton.

    Example:

        local = Local()

        some_cmd = local["some_cmd"]
        with local.cwd(some_path), local.env(FOO="bar"):
            some_cmd("baz")

        # A changed $PATH requires to rediscover commands.
        with local.prepend_path(important_path):
            local["other_cmd"]()
        with local.venv(venv_path):
            local["python"]("-m", "pip", "install", "salt")

    """

    def __init__(self):
        # Explicitly cast values to strings to avoid problems on Windows
        self._env = {k: str(v) for k, v in os.environ.items()}

    def __getitem__(self, exe):
        """
        Return a LocalCommand in this context.
        """
        return LocalCommand(exe, _local=self)

    @property
    def path(self):
        """
        List of paths in the context's $PATH.
        """
        return self._env.get("PATH", "").split(os.pathsep)

    @contextmanager
    def cwd(self, path):
        """
        Set the default current working directory for commands inside this context.
        """
        prev = Path(os.getcwd())
        new = prev / path
        os.cwd(new)
        try:
            yield
        finally:
            os.cwd(prev)

    @contextmanager
    def env(self, **kwargs):
        """
        Override default env vars (sourced from the current process' environment)
        for commands inside this context.
        """
        prev = self._env.copy()
        self._env.update((k, str(v)) for k, v in kwargs.items())
        try:
            yield
        finally:
            self._env = prev

    @contextmanager
    def path_prepend(self, *args):
        """
        Prepend paths to $PATH for commands inside this context.

        Note: If you have saved a reference to an already requested command,
        its $PATH will be updated, but it might not be the command
        that would have been returned by a new request.
        """
        new_path = [str(arg) for arg in args] + self.path
        with self.env(PATH=os.pathsep.join(new_path)):
            yield

    @contextmanager
    def venv(self, venv_dir):
        """
        Enter a Python virtual environment. Effectively prepends its bin dir
        to $PATH and sets ``VIRTUAL_ENV``.
        """
        venv_dir = Path(venv_dir)
        if not venv_dir.is_dir() or not (venv_dir / "pyvenv.cfg").exists():
            raise ValueError(f"Not a virtual environment: {venv_dir}")
        venv_bin_dir = venv_dir / "bin"
        if platform.system() == "Windows":
            venv_bin_dir = venv_dir / "Scripts"
        with self.path_prepend(venv_bin_dir), self.env(VIRTUAL_ENV=str(venv_dir)):
            yield


@dataclass(frozen=True)
class Executable:
    """
    Utility class used to avoid repeated command lookups.
    """

    _exe: str

    def __str__(self):
        return self._exe

    def __repr__(self):
        return f"Executable <{self._exe}>"


@dataclass(frozen=True)
class Command:
    """
    A command object, can be instantiated directly. Does not follow ``Local``.
    """

    exe: Executable | str
    args: tuple[str, ...] = ()

    def __post_init__(self):
        if not isinstance(self.exe, Executable):
            if not (full_exe := self._which(self.exe)):
                raise CommandNotFound(self.exe)
            object.__setattr__(self, "exe", Executable(full_exe))

    def _which(self, exe):
        return shutil.which(exe)

    def _get_env(self, overrides=None):
        base = {k: str(v) for k, v in os.environ.items()}
        base.update(overrides or {})
        return base

    def __getitem__(self, arg_or_args):
        """
        Returns a subcommand with bound parameters.

        Example:

            git = Command("git")["-c", "commit.gpgsign=0"]
            # ...
            git("add", ".")
            git("commit", "-m", "testcommit")

        """
        if not isinstance(arg_or_args, tuple):
            arg_or_args = (arg_or_args,)
        return type(self)(self.exe, tuple(*self.args, *arg_or_args), _local=self._local)

    def __call__(self, *args, **kwargs):
        """
        Run this command and return stdout.
        """
        return self.run(*args, **kwargs).stdout

    def __str__(self):
        return shlex.join([self.exe] + list(self.args))

    def __repr__(self):
        return f"Command<{self.exe}, {self.args!r}>"

    def run(self, *args, check=True, env=None, **kwargs):
        """
        Run this command and return the full output.
        """
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.PIPE)
        kwargs.setdefault("text", True)
        argv = [str(self.exe), *self.args, *args]
        proc = subprocess.run(argv, check=False, env=self._get_env(env), **kwargs)
        ret = ProcessResult(
            retcode=proc.returncode,
            stdout=proc.stdout,
            stderr=proc.stderr,
            argv=argv,
        )
        if check:
            ret.check()
        return ret


# Should be imported from here.
local = Local()


@dataclass(frozen=True)
class LocalCommand(Command):
    """
    Command returned by Local()["some_command"]. Follows local contexts.
    """

    _local: Local = field(kw_only=True, repr=False, default=local)

    def _which(self, exe):
        return shutil.which(exe, path=self._local._env.get("PATH", ""))

    def _get_env(self, overrides=None):
        base = self._local._env.copy()
        base.update(overrides or {})
        return base


# We must assume git is installed
git = local["git"]



from unittest.mock import Mock
from unittest.mock import patch

import pytest

from saltext.vault.modules import vault as vaultexe
from saltext.vault.states import vault


@pytest.fixture
def configure_loader_modules():
    return {vault: {}}


@pytest.fixture
def policy_fetch():
    fetch = Mock(return_value="test-rules", spec=vaultexe.policy_fetch)
    with patch.dict(vault.__salt__, {"vault.policy_fetch": fetch}):
        yield fetch


@pytest.fixture
def policy_write():
    write = Mock(return_value=True, spec=vaultexe.policy_write)
    with patch.dict(vault.__salt__, {"vault.policy_write": write}):
        yield write


@pytest.mark.usefixtures("policy_fetch")
@pytest.mark.parametrize("test", [False, True])
def test_policy_present_no_changes(test):
    """
    Test that when a policy is present as requested, no changes
    are reported for success, regardless of opts["test"].
    """
    with patch.dict(vault.__opts__, {"test": test}):
        res = vault.policy_present("test-policy", "test-rules")
    assert res["result"]
    assert not res["changes"]


@pytest.mark.parametrize("test", [False, True])
def test_policy_present_create(policy_fetch, policy_write, test):
    """
    Test that when a policy does not exist, it will be created.
    The function should respect opts["test"].
    """
    policy_fetch.return_value = None
    with patch.dict(vault.__opts__, {"test": test}):
        res = vault.policy_present("test-policy", "test-rules")
    assert res["changes"]
    if test:
        assert res["result"] is None
        assert "would be created" in res["comment"]
        policy_write.assert_not_called()
    else:
        assert res["result"]
        assert "has been created" in res["comment"]
        policy_write.assert_called_once_with("test-policy", "test-rules")


@pytest.mark.usefixtures("policy_fetch")
@pytest.mark.parametrize("test", [False, True])
def test_policy_present_changes(policy_write, test):
    """
    Test that when a policy exists, but the rules need to be updated,
    it is detected and respects the value of opts["test"].
    """
    with patch.dict(vault.__opts__, {"test": test}):
        res = vault.policy_present("test-policy", "new-test-rules")
    assert res["changes"]
    if test:
        assert res["result"] is None
        assert "would be updated" in res["comment"]
        policy_write.assert_not_called()
    else:
        assert res["result"]
        assert "has been updated" in res["comment"]
        policy_write.assert_called_once_with("test-policy", "new-test-rules")


@pytest.mark.parametrize("test", [False, True])
def test_policy_absent_no_changes(policy_fetch, test):
    """
    Test that when a policy is absent as requested, no changes
    are reported for success, regardless of opts["test"].
    """
    policy_fetch.return_value = None
    with patch.dict(vault.__opts__, {"test": test}):
        res = vault.policy_absent("test-policy")
    assert res["result"]
    assert not res["changes"]


@pytest.mark.usefixtures("policy_fetch")
@pytest.mark.parametrize("test", [False, True])
def test_policy_absent_changes(test):
    """
    Test that when a policy exists, it will be deleted.
    The function should respect opts["test"].
    """
    delete = Mock(spec=vaultexe.policy_delete)
    with patch.dict(vault.__salt__, {"vault.policy_delete": delete}):
        with patch.dict(vault.__opts__, {"test": test}):
            res = vault.policy_absent("test-policy")
        assert res["changes"]
        if test:
            assert res["result"] is None
            assert "would be deleted" in res["comment"]
            delete.assert_not_called()
        else:
            assert res["result"]
            assert "has been deleted" in res["comment"]
            delete.assert_called_once_with("test-policy")



from datetime import datetime
from datetime import timedelta
from datetime import timezone

import pytest
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from salt.utils.x509 import generate_rsa_privkey
from salt.utils.x509 import load_cert

from saltext.vault.utils.vault.pki import dec2hex
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_list_detailed
from tests.support.vault import vault_read
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "features": {
            "x509_v2": True,
        },
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "cache": {
                "backend": "disk",  # ensure a persistent cache is available for get_creds
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
    }


@pytest.fixture
def testrole():
    return {"ttl": 3600, "max_ttl": 86400, "allow_any_name": True, "issuer_ref": "testissuer"}


@pytest.fixture
def testissuer():
    return {"issuer_name": "testissuer", "common_name": "Test Issuer CA"}


@pytest.fixture
def testissuer2():
    return {"issuer_name": "testissuer2", "common_name": "Test Issuer CA 2"}


@pytest.fixture(scope="module")
def private_key():
    pk = generate_rsa_privkey(2048)
    pk_bytes = pk.private_bytes(
        serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return pk_bytes.decode()


@pytest.fixture(scope="module", autouse=True)
def pki_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("pki")
    yield
    assert vault_disable_secret_engine("pki")


@pytest.fixture(params=[["testrole"]])
def roles_setup(request):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            vault_write(f"pki/roles/{role_name}", **role_args)
            assert role_name in vault_list("pki/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("pki/roles"):
                vault_delete(f"pki/roles/{role_name}")
                assert role_name not in vault_list("pki/roles")


@pytest.fixture
def root_issuer_setup():
    root_ca_args = {"issuer_name": "test-issuer-root", "common_name": "Test Issuer Root CA"}

    vault_write("pki/root/generate/internal", **root_ca_args)
    assert vault_read(f"pki/issuer/{root_ca_args['issuer_name']}")


@pytest.fixture(params=[["testissuer"]])
def issuers_setup(request, root_issuer_setup):  # pylint: disable=unused-argument
    try:
        for issuer_name in request.param:
            issuer_args = request.getfixturevalue(issuer_name)

            csr_resp = vault_write("pki/intermediate/generate/internal", **issuer_args)["data"]
            sign_resp = vault_write(
                "/pki/root/sign-intermediate", csr=csr_resp["csr"], **issuer_args
            )["data"]
            resp = vault_write(
                "/pki/intermediate/set-signed", certificate=sign_resp["certificate"]
            )["data"]
            for issuer in resp["imported_issuers"]:
                vault_write(f"/pki/issuer/{issuer}", **issuer_args)
                assert vault_read(f"pki/issuer/{issuer_name}")
        yield
    finally:
        all_issuers = vault_list_detailed("pki/issuers")
        issuers_names = []
        if len(all_issuers) > 0:
            issuers_names = [v["issuer_name"] for k, v in all_issuers["key_info"].items()]
        for issuer_name in request.param:
            if issuer_name in issuers_names:
                vault_delete(f"pki/issuer/{issuer_name}")


@pytest.fixture
def vault_pki(modules):
    try:
        yield modules.vault_pki
    finally:
        if "testrole" in vault_list("pki/roles"):
            vault_delete("pki/roles/testrole")
            assert "testrole" not in vault_list("pki/roles")

        vault_delete("pki/issuer/test-issuer-root")


@pytest.fixture
def generated_root():
    yield
    vault_delete("pki/issuer/generated-root")


@pytest.mark.usefixtures("roles_setup")
def test_list_roles(vault_pki):
    ret = vault_pki.list_roles()
    assert ret == ["testrole"]


def test_list_empty_roles(vault_pki):
    ret = vault_pki.list_roles()
    assert ret == []


@pytest.mark.usefixtures("roles_setup")
def test_delete_role(vault_pki):
    ret = vault_pki.delete_role("testrole")
    assert ret
    assert "testrole" not in vault_list("pki/roles")


def test_write_role(vault_pki):
    assert vault_pki.write_role("testrole2", ttl="360h") is True
    assert "testrole2" in vault_list("pki/roles")


@pytest.mark.usefixtures("roles_setup")
def test_update_role(vault_pki):
    assert vault_pki.write_role("testrole", ttl="4h", max_ttl="24h") is True
    assert vault_read("pki/roles/testrole")["data"]["ttl"] == 14400  # 4 hours in seconds


@pytest.mark.usefixtures("issuers_setup")
def test_list_issuers(vault_pki):
    ret = [info["issuer_name"] for info in vault_pki.list_issuers().values()]
    assert set(ret) == {"test-issuer-root", "testissuer"}


@pytest.mark.usefixtures("issuers_setup")
def test_read_issuer(vault_pki):
    ret = vault_pki.read_issuer("testissuer")
    assert ret["issuer_name"] == "testissuer"
    assert "ca_chain" in ret
    assert "certificate" in ret


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_issue_certificate(vault_pki):
    # Certificate expiration is always in UTC, so we need to compare with UTC.
    run_time = datetime.now(tz=timezone.utc)

    ret = vault_pki.issue_certificate(
        role_name="testrole",
        common_name="test.example.com",
        ttl="2h",
        alt_names=["DNS:test2.example.com"],
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])

    assert certificate.issuer.rfc4514_string() == "CN=Test Issuer CA"

    assert certificate.subject.rfc4514_string() == "CN=test.example.com"
    san = certificate.extensions.get_extension_for_class(x509.SubjectAlternativeName)
    dns_sans = san.value.get_values_for_type(x509.DNSName)
    assert certificate.not_valid_after_utc - run_time > timedelta(hours=1)
    assert "test2.example.com" in dns_sans
    assert "test.example.com" in dns_sans


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_sign_certificate_with_private_key(vault_pki, private_key):
    # Certificate expiration is always in UTC, so we need to compare with UTC.
    run_time = datetime.now(tz=timezone.utc)

    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
        ttl="2h",
        alt_names=["dns:test2.example.com"],
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])

    assert certificate.issuer.rfc4514_string() == "CN=Test Issuer CA"

    assert certificate.subject.rfc4514_string() == "CN=test.example.com"
    san = certificate.extensions.get_extension_for_class(x509.SubjectAlternativeName)
    dns_sans = san.value.get_values_for_type(x509.DNSName)
    assert certificate.not_valid_after_utc - run_time > timedelta(hours=1)
    assert "test2.example.com" in dns_sans
    assert "test.example.com" in dns_sans


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_sign_certificate_with_sign_verbatim(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
        ttl="2h",
        sign_verbatim=True,
        alt_names=["dns:test2.example.com"],
        L="Boston",
        C="US",
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])
    assert certificate.subject.get_attributes_for_oid(x509.OID_COUNTRY_NAME)[0].value == "US"
    assert certificate.subject.get_attributes_for_oid(x509.OID_LOCALITY_NAME)[0].value == "Boston"


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("issuers_setup", [["testissuer", "testissuer2"]], indirect=True)
def test_sign_certificate_with_alternative_issuer(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
        ttl="2h",
        sign_verbatim=True,
        issuer_ref="testissuer2",
        alt_names=["dns:test2.example.com"],
        L="Boston",
        C="US",
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])
    assert certificate.issuer.rfc4514_string() == "CN=Test Issuer CA 2"


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_sign_certificate_with_der_encoding(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
        ttl="2h",
        encoding="der",
        alt_names=["dns:test2.example.com"],
    )

    assert "certificate" in ret
    _, encoding, _, _ = load_cert(ret["certificate"], get_encoding=True)
    assert encoding.lower() == "der"


@pytest.mark.usefixtures("issuers_setup")
def test_read_issuer_certificate(vault_pki):
    ret = vault_pki.read_issuer_certificate("testissuer")

    certificate = load_cert(ret)

    assert certificate.subject.rfc4514_string() == "CN=Test Issuer CA"


@pytest.mark.usefixtures("issuers_setup")
def test_read_issuer_certificate_with_chain(vault_pki):
    ret = vault_pki.read_issuer_certificate("testissuer", include_chain=True)

    certificate, chain = load_cert(ret, load_chain=True)

    assert certificate.subject.rfc4514_string() == "CN=Test Issuer CA"
    assert len(chain) == 1

    assert chain[0].subject.rfc4514_string() == "CN=Test Issuer Root CA"


@pytest.mark.usefixtures("issuers_setup")
def test_delete_issuer(vault_pki):
    ret = [info["issuer_name"] for info in vault_pki.list_issuers().values()]
    assert "testissuer" in ret

    ret = vault_pki.delete_issuer("testissuer")
    assert ret

    ret = [info["issuer_name"] for info in vault_pki.list_issuers().values()]
    assert "testissuer" not in ret


@pytest.mark.usefixtures("issuers_setup")
def test_delete_issuer_with_private_key(vault_pki):
    ret = vault_pki.read_issuer("testissuer")
    assert ret["key_id"]
    private_key_id = ret["key_id"]

    ret = vault_pki.delete_issuer("testissuer", include_key=True)
    assert ret

    ret = [info["issuer_name"] for info in vault_pki.list_issuers().values()]
    assert "testissuer" not in ret

    keys = vault_list("pki/keys")
    assert private_key_id not in keys


@pytest.mark.usefixtures("issuers_setup")
def test_update_issuer(vault_pki):
    ret = vault_pki.update_issuer(
        "testissuer",
        aia_urls=["http://aia.example.com/ca.list"],
    )
    assert ret

    # Now update OCSP endpoints.
    ret = vault_pki.update_issuer(
        "testissuer",
        ocsp_servers=["http://ocsp.example.com"],
    )

    # Now update CRL endpoints.
    ret = vault_pki.update_issuer(
        "testissuer",
        crl_endpoints=["http://crl.example.com/ca.crl"],
    )

    ret = vault_pki.read_issuer("testissuer")
    assert "http://crl.example.com/ca.crl" in ret["crl_distribution_points"]
    assert "http://aia.example.com/ca.list" in ret["issuing_certificates"]
    assert "http://ocsp.example.com" in ret["ocsp_servers"]


@pytest.mark.usefixtures("issuers_setup")
def test_read_issuer_crl(vault_pki):
    ret_complete = vault_pki.read_issuer_crl("testissuer")

    crl_complete = x509.load_pem_x509_crl(ret_complete.encode())
    assert crl_complete.issuer.rfc4514_string() == "CN=Test Issuer CA"
    ret_delta = vault_pki.read_issuer_crl("testissuer", delta=True)
    crl_delta = x509.load_pem_x509_crl(ret_delta.encode())
    assert crl_delta.issuer.rfc4514_string() == "CN=Test Issuer CA"
    assert ret_delta != ret_complete


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_revoke_certificate(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])

    serial = dec2hex(certificate.serial_number)

    ret = vault_pki.revoke_certificate(serial=serial)
    assert ret

    revoked_certs = [serial.upper() for serial in vault_pki.list_revoked_certificates()]
    assert serial in revoked_certs


@pytest.mark.usefixtures("issuers_setup")
def test_get_default_issuer(vault_pki):
    before = vault_pki.get_default_issuer()
    assert before

    # Now, delete default issuer
    vault_pki.delete_issuer(before)

    after = vault_pki.get_default_issuer()
    assert after is None


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.parametrize("issuers_setup", [["testissuer", "testissuer2"]], indirect=True)
def test_set_default_issuer(vault_pki):
    before = vault_pki.get_default_issuer()

    assert before
    ret = vault_pki.set_default_issuer(name="testissuer2")
    assert ret

    after = vault_pki.get_default_issuer()
    assert after != before


@pytest.mark.usefixtures("generated_root")
def test_generate_root(vault_pki):
    ret = vault_pki.list_issuers()
    assert ret == []

    ret = vault_pki.generate_root(
        common_name="generated root",
        issuer_name="generated-root",
    )

    assert "certificate" in ret

    certificate = load_cert(ret["certificate"])
    assert certificate.subject.rfc4514_string() == "CN=generated root"

    ret = vault_pki.list_issuers()
    assert len(ret) == 1


@pytest.mark.usefixtures("generated_root")
def test_generate_root_exported(vault_pki):
    ret = vault_pki.list_issuers()
    assert ret == []

    ret = vault_pki.generate_root(
        common_name="generated root",
        issuer_name="generated-root",
        type="exported",
    )

    assert "certificate" in ret
    assert "private_key" in ret


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_list_certificates(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
    )
    assert "certificate" in ret
    certificate = load_cert(ret["certificate"])

    serial = dec2hex(certificate.serial_number)

    all_certs = [serial.upper() for serial in vault_pki.list_certificates()]

    assert serial in all_certs


@pytest.mark.usefixtures("issuers_setup")
@pytest.mark.usefixtures("roles_setup")
def test_read_certificate(vault_pki, private_key):
    ret = vault_pki.sign_certificate(
        "testrole",
        common_name="test.example.com",
        private_key=private_key,
    )
    assert "certificate" in ret

    signed_certificate = load_cert(ret["certificate"])
    serial = dec2hex(signed_certificate.serial_number)
    ret = vault_pki.read_certificate(serial)
    read_certificate = load_cert(ret)
    assert read_certificate.serial_number == signed_certificate.serial_number



from pathlib import Path

from .cmd import git


def ensure_git():
    """
    Ensure the repository has been initialized.
    """
    if Path(".git").is_dir():
        return
    git("init", "--initial-branch", "main")


def list_untracked():
    """
    List untracked files.
    """
    for path in git("ls-files", "-z", "-o", "--exclude-standard").split("\x00"):
        if path:
            yield path


def list_conflicted():
    """
    List files with merge conflicts.
    """
    for path in git("diff", "-z", "--name-only", "--diff-filter=U", "--relative").split("\x00"):
        if path:
            yield path



from pygments.lexer import bygroups
from pygments.lexer import inherit
from pygments.lexers.configs import TerraformLexer
from pygments.token import Keyword
from pygments.token import Name
from pygments.token import Punctuation
from pygments.token import Whitespace


class VaultPolicyLexer(TerraformLexer):
    aliases = ["vaultpolicy"]
    filenames = ["*.hcl"]
    mimetypes = ["application/x-hcl-policy"]

    tokens = {
        "basic": [
            inherit,
            (
                r"(path)(\s+)(\".*\")(\s+)(\{)",
                bygroups(Keyword.Reserved, Whitespace, Name.Variable, Whitespace, Punctuation),
            ),
        ],
    }


def setup(app):
    app.add_lexer("vaultpolicy", VaultPolicyLexer)
    return {"parallel_read_safe": True}



"""
Manage the Vault (or OpenBao) database secret engine, request and cache
leased database credentials.

.. versionadded:: 1.1.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging
from datetime import datetime
from datetime import timezone

from salt.defaults import NOT_SET
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils import vault
from saltext.vault.utils.vault import db as vaultdb

log = logging.getLogger(__name__)


def list_connections(mount="database"):
    """
    List configured database connections.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#list-connections>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.list_connections

    mount
        The mount path the DB backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/config"
    try:
        return vault.query("LIST", endpoint, __opts__, __context__)["data"]["keys"]
    except vault.VaultNotFoundError:
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def fetch_connection(name, mount="database"):
    """
    Read a configured database connection. Returns None if it does not exist.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#read-connection>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.fetch_connection mydb

    name
        The name of the database connection.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/config/{name}"
    try:
        return vault.query("GET", endpoint, __opts__, __context__)["data"]
    except vault.VaultNotFoundError:
        return None
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def write_connection(
    name,
    plugin,
    version="",
    verify=True,
    allowed_roles=None,
    root_rotation_statements=None,
    password_policy=None,
    rotate=True,
    mount="database",
    **kwargs,
):
    """
    Create/update a configured database connection.

    .. note::

        This endpoint distinguishes between create and update ACL capabilities.

    .. note::

        It is highly recommended to use a Vault-specific user rather than the admin user in the
        database when configuring the plugin. This user will be used to create/update/delete users
        within the database so it will need to have the appropriate permissions to do so.
        If the plugin supports rotating the root credentials, it is highly recommended to perform
        that action after configuring the plugin. This will change the password of the user
        configured in this step. The new password will not be viewable by users.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#configure-connection>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.write_connection mydb elasticsearch \
                url=http://127.0.0.1:9200 username=vault password=hunter2

    name
        The name of the database connection.

    plugin
        The name of the database plugin. Known plugins to this module are:
        ``cassandra``, ``couchbase``, ``elasticsearch``, ``influxdb``, ``hanadb``, ``mongodb``,
        ``mongodb_atlas``, ``mssql``, ``mysql``, ``oracle``, ``postgresql``, ``redis``,
        ``redis_elasticache``, ``redshift``, ``snowflake``.
        If you pass an unknown plugin, make sure its Vault-internal name can be formatted
        as ``{plugin}-database-plugin`` and to pass all required parameters as kwargs.

    version
        Specifies the semantic version of the plugin to use for this connection.

    verify
        Verify the connection during initial configuration. Defaults to True.

    allowed_roles
        List of the roles allowed to use this connection. ``["*"]`` means any role
        can use this connection. Defaults to empty (no role can use it).

    root_rotation_statements
        Specifies the database statements to be executed to rotate the root user's credentials.
        See the plugin's API page for more information on support and formatting for this parameter.

    password_policy
        The name of the password policy to use when generating passwords for this database.
        If not specified, this will use a default policy defined as:
        20 characters with at least 1 uppercase, 1 lowercase, 1 number, and 1 dash character.

    rotate
        Rotate the root credentials after plugin setup. Defaults to True.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.

    kwargs
        Different plugins require different parameters. You need to make sure that you pass them
        as supplemental keyword arguments. For known plugins, the required arguments will
        be checked.
    """
    endpoint = f"{mount}/config/{name}"
    plugin_meta = vaultdb.get_plugin_meta(plugin)
    plugin_name = plugin_meta["name"] or plugin
    payload = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    if fetch_connection(name, mount=mount) is None:
        missing_kwargs = set(plugin_meta["required"]) - set(payload)
        if missing_kwargs:
            raise SaltInvocationError(
                f"The plugin {plugin} requires the following additional kwargs: {missing_kwargs}."
            )

    payload["plugin_name"] = f"{plugin_name}-database-plugin"
    payload["verify_connection"] = verify
    if version is not None:
        payload["plugin_version"] = version
    if allowed_roles is not None:
        payload["allowed_roles"] = allowed_roles
    if root_rotation_statements is not None:
        payload["root_rotation_statements"] = root_rotation_statements
    if password_policy is not None:
        payload["password_policy"] = password_policy

    try:
        vault.query("POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err

    if not rotate:
        return True
    return rotate_root(name, mount=mount)


def delete_connection(name, mount="database"):
    """
    Delete a configured database connection. Returns None if it does not exist.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#delete-connection>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.delete_connection mydb

    name
        The name of the database connection.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/config/{name}"
    try:
        return vault.query("DELETE", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def reset_connection(name, mount="database"):
    """
    Close a connection and restart its plugin with the configuration stored in the barrier.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#reset-connection>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.reset_connection mydb

    name
        The name of the database connection.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/reset/{name}"
    try:
        return vault.query("POST", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def rotate_root(name, mount="database"):
    """
    Rotate the "root" user credentials stored for the database connection.

    .. warning::

        The rotated password will not be accessible, so it is highly recommended to create
        a dedicated user account as Vault's configured "root".

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#rotate-root-credentials>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.rotate_root mydb

    name
        The name of the database connection.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/rotate-root/{name}"
    try:
        return vault.query("POST", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def list_roles(static=False, mount="database"):
    """
    List configured database roles.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#list-roles>`__.
    `API method docs static <https://developer.hashicorp.com/vault/api-docs/secret/databases#list-static-roles>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.list_roles

    static
        Whether to list static roles. Defaults to False.

    mount
        The mount path the DB backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/{'static-' if static else ''}roles"
    try:
        return vault.query("LIST", endpoint, __opts__, __context__)["data"]["keys"]
    except vault.VaultNotFoundError:
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def fetch_role(name, static=False, mount="database"):
    """
    Read a configured database role. Returns None if it does not exist.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#read-role>`__.
    `API method docs static <https://developer.hashicorp.com/vault/api-docs/secret/databases#read-static-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.fetch_role myrole

    name
        The name of the database role.

    static
        Whether this role is static. Defaults to False.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/{'static-' if static else ''}roles/{name}"
    try:
        return vault.query("GET", endpoint, __opts__, __context__)["data"]
    except vault.VaultNotFoundError:
        return None
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def write_static_role(
    name,
    connection,
    username,
    rotation_period,
    rotation_statements=None,
    credential_type=None,
    credential_config=None,
    mount="database",
):
    """
    Create/update a database Static Role. Mind that not all databases support Static Roles.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#create-static-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.write_static_role myrole mydb myuser 24h

    name
        The name of the database role.

    connection
        The name of the database connection this role applies to.

    username
        The username to manage.

    rotation_period
        Specifies the amount of time Vault should wait before rotating the password.
        The minimum is ``5s``.

    rotation_statements
        Specifies the database statements to be executed to rotate the password for the
        configured database user. Not every plugin type will support this functionality.

    credential_type
        Specifies the type of credential that will be generated for the role.
        Options include: ``password``, ``rsa_private_key``. Defaults to ``password``.
        See the plugin's API page for credential types supported by individual databases.

    credential_config
        Specifies the configuration for the given ``credential_type`` as a mapping.
        For ``password``, only ``password_policy`` can be passed.
        For ``rsa_private_key``, ``key_bits`` (defaults to 2048) and ``format``
        (defaults to ``pkcs8``) are available.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    payload = {
        "username": username,
        "rotation_period": rotation_period,
    }
    if rotation_statements is not None:
        payload["rotation_statements"] = rotation_statements
    return _write_role(
        name,
        connection,
        payload,
        credential_type=credential_type,
        credential_config=credential_config,
        static=True,
        mount=mount,
    )


def write_role(
    name,
    connection,
    creation_statements,
    default_ttl=None,
    max_ttl=None,
    revocation_statements=None,
    rollback_statements=None,
    renew_statements=None,
    credential_type=None,
    credential_config=None,
    mount="database",
):
    r"""
    Create/update a regular database role.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#create-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.write_role myrole mydb \
                \["CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}'", "GRANT SELECT ON *.* TO '{{name}}'@'%'"\]

    name
        The name of the database role.

    connection
        The name of the database connection this role applies to.

    creation_statements
        Specifies a list of database statements executed to create and configure a user,
        usually templated with {{name}} and {{password}}. Required.

    default_ttl
        Specifies the TTL for the leases associated with this role. Accepts time suffixed
        strings (1h) or an integer number of seconds. Defaults to system/engine default TTL time.

    max_ttl
        Specifies the maximum TTL for the leases associated with this role. Accepts time suffixed
        strings (1h) or an integer number of seconds. Defaults to sys/mounts's default TTL time;
        this value is allowed to be less than the mount max TTL (or, if not set,
        the system max TTL), but it is not allowed to be longer.

    revocation_statements
        Specifies a list of database statements to be executed to revoke a user.

    rollback_statements
        Specifies a list of database statements to be executed to rollback a create operation
        in the event of an error. Availability and formatting depend on the specific plugin.

    renew_statements
        Specifies a list of database statements to be executed to renew a user.
        Availability and formatting depend on the specific plugin.

    credential_type
        Specifies the type of credential that will be generated for the role.
        Options include: ``password``, ``rsa_private_key``. Defaults to ``password``.
        See the plugin's API page for credential types supported by individual databases.

    credential_config
        Specifies the configuration for the given ``credential_type`` as a mapping.
        For ``password``, only ``password_policy`` can be passed.
        For ``rsa_private_key``, ``key_bits`` (defaults to 2048) and ``format``
        (defaults to ``pkcs8``) are available.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    payload = {
        "creation_statements": creation_statements,
    }
    if default_ttl is not None:
        payload["default_ttl"] = default_ttl
    if max_ttl is not None:
        payload["max_ttl"] = max_ttl
    if revocation_statements is not None:
        payload["revocation_statements"] = revocation_statements
    if rollback_statements is not None:
        payload["rollback_statements"] = rollback_statements
    if renew_statements is not None:
        payload["renew_statements"] = renew_statements
    return _write_role(
        name,
        connection,
        payload,
        credential_type=credential_type,
        credential_config=credential_config,
        static=False,
        mount=mount,
    )


def _write_role(
    name,
    connection,
    payload,
    credential_type=None,
    credential_config=None,
    static=False,
    mount="database",
):
    endpoint = f"{mount}/{'static-' if static else ''}roles/{name}"
    payload["db_name"] = connection
    if credential_type is not None:
        payload["credential_type"] = credential_type
    if credential_config is not None:
        valid_cred_configs = {
            "password": ["password_policy"],
            "rsa_private_key": ["key_bits", "format"],
        }
        credential_type = credential_type or "password"
        if credential_type in valid_cred_configs:
            invalid_configs = set(credential_config) - set(valid_cred_configs[credential_type])
            if invalid_configs:
                raise SaltInvocationError(
                    f"The following options are invalid for credential type {credential_type}: {invalid_configs}"
                )
        payload["credential_config"] = credential_config
    try:
        return vault.query(
            "POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True
        )
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def delete_role(name, static=False, mount="database"):
    """
    Delete a configured database role.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#delete-role>`__.
    `API method docs static <https://developer.hashicorp.com/vault/api-docs/secret/databases#delete-static-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.delete_role myrole

    name
        The name of the database role.

    static
        Whether this role is static. Defaults to False.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/{'static-' if static else ''}roles/{name}"
    try:
        return vault.query("DELETE", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def get_creds(
    name,
    static=False,
    cache=True,
    valid_for=NOT_SET,
    check_server=False,
    renew_increment=NOT_SET,
    revoke_delay=NOT_SET,
    meta=NOT_SET,
    mount="database",
    _warn_about_attr_change=True,
):
    """
    Read credentials based on the named role.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/databases#generate-credentials>`__.
    `API method docs static <https://developer.hashicorp.com/vault/api-docs/secret/databases#get-static-credentials>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.get_creds myrole

    name
        The name of the database role.

    static
        Whether this role is static. Defaults to False.

    cache
        Whether to use cached credentials local to this minion to avoid
        unnecessary reissuance.
        When ``static`` is false, set this to a string to be able to use multiple
        distinct credentials using the same role on the same minion.
        Set this to false to disable caching.
        Defaults to true.

        .. note::

            This uses the same cache backend as the Vault integration, so make
            sure you configure a persistent backend like ``disk`` if you expect
            the credentials to survive a single run.

        .. hint::

            For some applications, you might need to know the resulting cache key
            of the credential lease. The cache key is composed as a concatenation
            of the following parts (with ``.``):

            * ``db``
            * ``<mount>``
            * ``dynamic`` if static=False, else ``static``
            * ``<name>`` (the name of the database role)
            * ``default`` if cache=True or static=True, otherwise the value of ``cache``

            Examples for a role named ``foo``:

            * mount=database, static=False, cache=True: ``db.database.dynamic.foo.default``
            * mount=mariadb, static=False, cache=alt: ``db.mariadb.dynamic.foo.alt``
            * mount=mariadb, static=True, cache=True: ``db.mariadb.static.foo.default``


    valid_for
        When using cache, ensure the credentials are valid for at least this
        amount of time, otherwise request new ones.
        This can be an integer, which will be interpreted as seconds, or a time string
        using the same format as Vault does:
        Suffix ``s`` for seconds, ``m`` for minuts, ``h`` for hours, ``d`` for days.
        This will be cached together with the lease and might be used by other
        modules later.

    check_server
        Check on the Vault server whether the lease is still active and was not
        revoked early. Defaults to false.

    renew_increment
        When using cache and ``valid_for`` results in a renewal attempt, request this
        amount of time extension on the lease. This will be cached together with the
        lease and might be used by other modules later.

    revoke_delay
        When using cache and ``valid_for`` results in a revocation, set the lease
        validity to this value to allow a short amount of delay between the issuance
        of the new lease and the revocation of the old one. Defaults to ``60``.
        This will be cached together with the lease and might be used by other
        modules later.

    meta
        When using cache, this value will be cached together with the lease. It will
        be emitted by the ``vault_lease`` beacon module whenever a lease is
        running out (usually because it cannot be extended further). It is intended
        to support the reactor in deciding what needs to be done in order
        to to reconfigure dependent, Vault-unaware software with newly issued
        credentials. Entirely optional.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/{'static-' if static else ''}creds/{name}"

    if cache:
        ckey = f"db.{mount}.{'static' if static else 'dynamic'}.{name}"
        if not static and isinstance(cache, str):
            ckey += f".{cache}"
        else:
            ckey += ".default"
        creds_cache = vault.get_lease_store(__opts__, __context__)
        cached_creds = creds_cache.get(
            ckey,
            valid_for=valid_for if valid_for is not NOT_SET else None,
            revoke=revoke_delay if revoke_delay is not NOT_SET else None,
            check_server=check_server,
        )
        if cached_creds:
            changed = False
            for attr, val in (
                ("min_ttl", valid_for),
                ("renew_increment", renew_increment),
                ("revoke_delay", revoke_delay),
                ("meta", meta),
            ):
                if val is not NOT_SET and getattr(cached_creds, attr) != val:
                    setattr(cached_creds, attr, val)
                    changed = True
            if changed:
                # Warn about changes if a lease is managed by the state module
                # and this function is called e.g. during YAML rendering, overwriting
                # the desired attributes. The state module sets this to false.
                if _warn_about_attr_change:
                    log.warning(f"Cached credential `{ckey}` changed lifecycle attributes")
                creds_cache.store(ckey, cached_creds)
            return cached_creds.data

    try:
        res = vault.query("GET", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err

    lease = vault.VaultLease(
        min_ttl=valid_for if valid_for is not NOT_SET else None,
        renew_increment=renew_increment if renew_increment is not NOT_SET else None,
        revoke_delay=revoke_delay if revoke_delay is not NOT_SET else None,
        meta=meta if meta is not NOT_SET else None,
        **res,
    )
    if cache:
        creds_cache.store(ckey, lease)
    return lease.data


def clear_cached(name=None, mount=None, cache=None, static=None, delta=None, flush_on_failure=True):
    """
    Clear and revoke cached database credentials matching specified parameters.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.clear_cached name=myrole mount=database
            salt '*' vault_db.clear_cached mount=database
            salt '*' vault_db.clear_cached

    name
        Only clear credentials using this role name.

    mount
        Only clear credentials from this mount.

    cache
        Only clear credentials using this cache name (refer to get_creds for details).

    static
        Only clear static (``True``) or dynamic (``False``) credentials.

    delta
        Time after which the leases should be revoked by Vault.
        Defaults to what was set on the lease(s) during creation or 60s.

    flush_on_failure
        If a revocation fails, remove the lease from cache anyways.
        Defaults to true.
    """
    creds_cache = vault.get_lease_store(__opts__, __context__)
    return creds_cache.revoke_cached(
        match=vaultdb.create_cache_pattern(name=name, mount=mount, cache=cache, static=static),
        delta=delta,
        flush_on_failure=flush_on_failure,
    )


def list_cached(name=None, mount=None, cache=None, static=None):
    """
    List cached database credentials matching specified parameters.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.list_cached name=myrole mount=database
            salt '*' vault_db.list_cached mount=database
            salt '*' vault_db.list_cached

    name
        Only list credentials using this role name.

    mount
        Only list credentials from this mount.

    cache
        Only list credentials using this cache name (refer to get_creds for details).

    static
        Only list static (``True``) or dynamic (``False``) credentials.
    """
    creds_cache = vault.get_lease_store(__opts__, __context__)
    info = creds_cache.list_info(
        match=vaultdb.create_cache_pattern(name=name, mount=mount, cache=cache, static=static)
    )
    for lease in info.values():
        for val in ("creation_time", "expire_time"):
            if val in lease:
                lease[val] = (
                    datetime.fromtimestamp(lease[val], tz=timezone.utc)
                    .astimezone()
                    .strftime("%Y-%m-%d %H:%M:%S %Z")
                )
    return info


def renew_cached(name=None, mount=None, cache=None, static=None, increment=None):
    """
    Renew cached database credentials matching specified parameters.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.renew_cached name=myrole mount=database
            salt '*' vault_db.renew_cached mount=database
            salt '*' vault_db.renew_cached

    name
        Only renew credentials using this role name.

    mount
        Only renew credentials from this mount.

    cache
        Only renew credentials using this cache name (refer to get_creds for details).

    static
        Only renew static (``True``) or dynamic (``False``) credentials.

    increment
        Request the leases to be valid for this amount of time from the current
        point of time onwards. Can also be used to reduce the validity period.
        The server might not honor this increment.
        Can be an integer (seconds) or a time string like ``1h``. Optional.
        If unset, defaults to what was set on the lease during creation or
        the lease's default TTL.
    """
    creds_cache = vault.get_lease_store(__opts__, __context__)
    return creds_cache.renew_cached(
        match=vaultdb.create_cache_pattern(name=name, mount=mount, cache=cache, static=static),
        increment=increment,
    )


def rotate_static_role(name, mount="database"):
    """
    Rotate Static Role credentials stored for a given role name.

    `API method docs static <https://developer.hashicorp.com/vault/api-docs/secret/databases#rotate-static-role-credentials>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_db.rotate_static_role mystaticrole

    name
        The name of the database role.

    mount
        The mount path the database backend is mounted to. Defaults to ``database``.
    """
    endpoint = f"{mount}/rotate-role/{name}"
    try:
        return vault.query("POST", endpoint, __opts__, __context__)
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err



"""
Tests for the Vault module
"""

import logging
import shutil
import time

import pytest
from saltfactories.utils import random_string

from tests.support.vault import vault_delete_secret
from tests.support.vault import vault_list_secrets
from tests.support.vault import vault_write_secret

pytest.importorskip("docker")

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="class")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="class")
def vault_salt_master(
    salt_factories, pillar_state_tree, vault_port, vault_master_config
):  # pylint: disable=unused-argument
    factory = salt_factories.salt_master_daemon("vault-exemaster", defaults=vault_master_config)
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def vault_salt_minion(vault_salt_master, vault_minion_config):
    assert vault_salt_master.is_running()
    factory = vault_salt_master.salt_minion_daemon(
        random_string("vault-exeminion", uppercase=False),
        defaults=vault_minion_config,
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def vault_minion_config():
    return {"open_mode": True, "grains": {}}


@pytest.fixture(scope="class")
def vault_salt_run_cli(vault_salt_master):
    return vault_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def vault_salt_call_cli(vault_salt_minion):
    return vault_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def pillar_dual_use_tree(
    vault_salt_master,
    vault_salt_minion,
):
    top_pillar_contents = f"""
    base:
      '{vault_salt_minion.id}':
        - testvault
    """
    test_pillar_contents = """
    test:
      foo: bar
      jvmdump_pubkey: {{ salt["vault.read_secret"]("secret/test/jvmdump/ssh_key", "public_key") }}
      jenkins_pubkey: {{ salt["vault.read_secret"]("secret/test/jenkins/master/ssh_key", "public_key") }}
    """
    top_file = vault_salt_master.pillar_tree.base.temp_file("top.sls", top_pillar_contents)
    test_file = vault_salt_master.pillar_tree.base.temp_file("testvault.sls", test_pillar_contents)

    with top_file, test_file:
        yield


@pytest.fixture(scope="class")
def vault_testing_data(container):  # pylint: disable=unused-argument
    vault_write_secret("secret/test/jvmdump/ssh_key", public_key="yup_dump")
    vault_write_secret("secret/test/jenkins/master/ssh_key", public_key="yup_master")
    vault_write_secret("secret/test/deleteme", pls=":)")
    try:
        yield
    finally:
        vault_delete_secret("secret/test/jvmdump/ssh_key")
        vault_delete_secret("secret/test/jenkins/master/ssh_key")
        for x in ("deleteme", "write"):
            if x in vault_list_secrets("secret/test"):
                vault_delete_secret(f"secret/test/{x}")


@pytest.mark.usefixtures("vault_testing_data", "pillar_dual_use_tree")
class TestSingleUseToken:
    @pytest.fixture(scope="class")
    def vault_master_config(self, pillar_state_tree, vault_port):
        return {
            "pillar_roots": {"base": [str(pillar_state_tree)]},
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_new_token",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "token",
                    "token": {
                        "params": {
                            "num_uses": 1,
                        }
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                    ]
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
            "minion_data_cache": False,
        }

    def test_vault_read_secret(self, vault_salt_call_cli):
        """
        Test that the Vault module can fetch a single secret when tokens
        are issued with uses=1.
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/test/jvmdump/ssh_key")
        assert ret.returncode == 0
        assert ret.data == {"public_key": "yup_dump"}

    def test_vault_read_secret_can_fetch_more_than_one_secret_in_one_run(
        self,
        vault_salt_call_cli,
        caplog,
    ):
        """
        Test that the Vault module can fetch multiple secrets during
        a single run when tokens are issued with uses=1.
        Issue #57561
        """
        ret = vault_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        ret = vault_salt_call_cli.run("pillar.items")
        assert ret.returncode == 0
        assert ret.data
        assert "Pillar render error" not in caplog.text
        assert "test" in ret.data
        assert "jvmdump_pubkey" in ret.data["test"]
        assert ret.data["test"]["jvmdump_pubkey"] == "yup_dump"
        assert "jenkins_pubkey" in ret.data["test"]
        assert ret.data["test"]["jenkins_pubkey"] == "yup_master"

    def test_vault_write_secret(self, vault_salt_call_cli):
        """
        Test that the Vault module can write a single secret when tokens
        are issued with uses=1.
        """
        ret = vault_salt_call_cli.run("vault.write_secret", "secret/test/write", success="yup")
        assert ret.returncode == 0
        assert ret.data
        assert "write" in vault_list_secrets("secret/test")

    def test_vault_delete_secret(self, vault_salt_call_cli):
        """
        Test that the Vault module can delete a single secret when tokens
        are issued with uses=1.
        """
        ret = vault_salt_call_cli.run("vault.delete_secret", "secret/test/deleteme")
        assert ret.returncode == 0
        assert ret.data
        assert "delete" not in vault_list_secrets("secret/test")


class TestTokenMinimumTTLUnrenewable:
    """
    Test that a new token is requested when the current one does not
    fulfill minimum_ttl and cannot be renewed
    """

    @pytest.fixture(scope="class")
    def vault_master_config(self, vault_port):
        return {
            "pillar_roots": {},
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_new_token",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "token",
                    "token": {
                        "params": {
                            "num_uses": 0,
                            "explicit_max_ttl": 180,
                        }
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                    ]
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
            "minion_data_cache": False,
        }

    @pytest.fixture(scope="class")
    def vault_minion_config(self):
        return {
            "open_mode": True,
            "grains": {},
            "vault": {"auth": {"token_lifecycle": {"minimum_ttl": 178, "renew_increment": None}}},
        }

    def test_minimum_ttl_is_respected(self, vault_salt_call_cli):
        # create token by looking it up
        ret = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret.data
        assert ret.returncode == 0
        # wait
        time_before = time.time()
        while time.time() - time_before < 3:
            time.sleep(0.1)
        # reissue token by looking it up
        ret_new = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret_new.returncode == 0
        assert ret_new.data
        # ensure a new token was created, even though the previous one would have been
        # valid still
        assert ret_new.data["data"]["id"] != ret.data["data"]["id"]


class TestTokenMinimumTTLRenewable:
    """
    Test that tokens are renewed and minimum_ttl is respected
    """

    @pytest.fixture(scope="class")
    def vault_master_config(self, vault_port):
        return {
            "pillar_roots": {},
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_new_token",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "token",
                    "token": {
                        "params": {
                            "num_uses": 0,
                            "ttl": 180,
                        }
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                    ]
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
            "minion_data_cache": False,
        }

    @pytest.fixture(scope="class")
    def vault_minion_config(self):
        return {
            "open_mode": True,
            "grains": {},
            "vault": {"auth": {"token_lifecycle": {"minimum_ttl": 177, "renew_increment": None}}},
        }

    def test_minimum_ttl_is_respected(self, vault_salt_call_cli):
        # create token by looking it up
        ret = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret.data
        assert ret.returncode == 0
        # wait
        time_before = time.time()
        while time.time() - time_before < 4:
            time.sleep(0.1)
        # renew token by looking it up
        ret_new = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret_new.returncode == 0
        assert ret_new.data
        # ensure the current token's validity has been extended
        assert ret_new.data["data"]["id"] == ret.data["data"]["id"]
        assert ret_new.data["data"]["expire_time"] > ret.data["data"]["expire_time"]



import logging
import time
from datetime import datetime

import pytest
from saltfactories.utils import random_string

from tests.support.mysql import MySQLImage
from tests.support.mysql import create_mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_container  # pylint: disable=unused-import
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_revoke
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "cache": {
                "backend": "disk",  # ensure a persistent cache is available for get_creds
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture(scope="module")
def mysql_image():
    version = "10.3"
    return MySQLImage(
        name="mariadb",
        tag=version,
        container_id=random_string(f"mariadb-{version}-"),
    )


@pytest.fixture
def role_args_common():
    return {
        "db_name": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def testrole():
    return {
        "default_ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def testreissuerole():
    return {
        "default_ttl": 180,
        "max_ttl": 180,
    }


@pytest.fixture
def teststaticrole(mysql_container):
    return {
        "db_name": "testdb",
        "rotation_period": 86400,
        "username": mysql_container.mysql_user,
    }


@pytest.fixture
def testdb(mysql_container, container_host_ref):
    return {
        "plugin_name": "mysql-database-plugin",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": "testrole,teststaticrole,testreissuerole",
        "username": "root",
        "password": mysql_container.mysql_passwd,
    }


@pytest.fixture(scope="module", autouse=True)
def db_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("database")
    yield
    assert vault_disable_secret_engine("database")


@pytest.fixture
def connection_setup(testdb):
    try:
        vault_write("database/config/testdb", **testdb)
        assert "testdb" in vault_list("database/config")
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")


@pytest.fixture(params=[["testrole"]])
def roles_setup(connection_setup, request, role_args_common):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            role_args.update(role_args_common)
            vault_write(f"database/roles/{role_name}", **role_args)
            assert role_name in vault_list("database/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("database/roles"):
                vault_delete(f"database/roles/{role_name}")
                assert role_name not in vault_list("database/roles")


@pytest.fixture
def role_static_setup(connection_setup, teststaticrole):  # pylint: disable=unused-argument
    role_name = "teststaticrole"
    try:
        vault_write(f"database/static-roles/{role_name}", **teststaticrole)
        assert role_name in vault_list("database/static-roles")
        yield
    finally:
        if role_name in vault_list("database/static-roles"):
            vault_delete(f"database/static-roles/{role_name}")
            assert role_name not in vault_list("database/static-roles")


@pytest.fixture
def vault_db(modules):
    try:
        yield modules.vault_db
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")
        if "testrole" in vault_list("database/roles"):
            vault_delete("database/roles/testrole")
            assert "testrole" not in vault_list("database/roles")
        if "teststaticrole" in vault_list("database/static-roles"):
            vault_delete("database/static-roles/teststaticrole")
            assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.mark.usefixtures("connection_setup")
def test_list_connections(vault_db):
    ret = vault_db.list_connections()
    assert ret == ["testdb"]


def test_list_connections_empty(vault_db):
    ret = vault_db.list_connections()
    assert ret == []


@pytest.mark.usefixtures("connection_setup")
def test_fetch_connection(vault_db, testdb):
    ret = vault_db.fetch_connection("testdb")
    assert ret
    for var, val in testdb.items():
        if var == "password":
            continue
        if var in ["connection_url", "username"]:
            assert var in ret["connection_details"]
            assert ret["connection_details"][var] == val
        else:
            assert var in ret
            if var == "allowed_roles":
                assert ret[var] == list(val.split(","))
            else:
                assert ret[var] == val


@pytest.mark.usefixtures("testdb")
def test_fetch_connection_empty(vault_db):
    ret = vault_db.fetch_connection("foobar")
    assert ret is None


@pytest.mark.usefixtures("testdb")
def test_write_connection(vault_db, mysql_container, container_host_ref):
    args = {
        "plugin": "mysql",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": ["testrole", "teststaticrole"],
        "username": "root",
        "password": mysql_container.mysql_passwd,
        # Can't rotate because we wouldn't know the new one for further tests
        "rotate": False,
    }
    ret = vault_db.write_connection("testdb", **args)
    assert ret
    assert "testdb" in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_update_connection(vault_db):
    """
    Ensure missing kwargs are not enforced on updates.
    """
    assert vault_db.write_connection("testdb", "mysql", allowed_roles=["*"], rotate=False) is True


@pytest.mark.usefixtures("connection_setup")
def test_delete_connection(vault_db):
    ret = vault_db.delete_connection("testdb")
    assert ret
    assert "testdb" not in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_reset_connection(vault_db):
    ret = vault_db.reset_connection("testdb")
    assert ret


@pytest.mark.usefixtures("roles_setup")
def test_list_roles(vault_db):
    ret = vault_db.list_roles()
    assert ret == ["testrole"]


def test_list_roles_empty(vault_db):
    ret = vault_db.list_roles()
    assert ret == []


@pytest.mark.usefixtures("role_static_setup")
def test_list_roles_static(vault_db):
    ret = vault_db.list_roles(static=True)
    assert ret == ["teststaticrole"]


@pytest.mark.usefixtures("roles_setup")
def test_fetch_role(vault_db, testrole):
    ret = vault_db.fetch_role("testrole")
    assert ret
    for var, val in testrole.items():
        assert var in ret
        if var == "creation_statements":
            assert ret[var] == [val]
        else:
            assert ret[var] == val


@pytest.mark.usefixtures("role_static_setup")
def test_fetch_role_static(vault_db, teststaticrole):
    ret = vault_db.fetch_role("teststaticrole", static=True)
    assert ret
    for var, val in teststaticrole.items():
        assert var in ret
        assert ret[var] == val


def test_fetch_role_empty(vault_db):
    ret = vault_db.fetch_role("foobar")
    assert ret is None


@pytest.mark.usefixtures("connection_setup")
def test_write_role(vault_db):
    args = {
        "connection": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }
    ret = vault_db.write_role("testrole", **args)
    assert ret
    assert "testrole" in vault_list("database/roles")


@pytest.mark.usefixtures("connection_setup")
def test_write_static_role(vault_db, mysql_container):
    args = {
        "connection": "testdb",
        "username": mysql_container.mysql_user,
        "rotation_period": 86400,
    }
    ret = vault_db.write_static_role("teststaticrole", **args)
    assert ret
    assert "teststaticrole" in vault_list("database/static-roles")


@pytest.mark.usefixtures("roles_setup")
def test_delete_role(vault_db):
    ret = vault_db.delete_role("testrole")
    assert ret
    assert "testrole" not in vault_list("database/roles")


@pytest.mark.usefixtures("role_static_setup")
def test_delete_role_static(vault_db):
    ret = vault_db.delete_role("teststaticrole", static=True)
    assert ret
    assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.fixture(params=({},))
def _cached_creds(vault_db, roles_setup, request):  # pylint: disable=unused-argument
    data = request.param.copy()
    role = data.pop("role", "testrole")
    ret = vault_db.get_creds(role, cache=True, **data)
    assert ret
    assert "username" in ret
    assert "password" in ret
    return ret


@pytest.mark.usefixtures("roles_setup")
def test_get_creds(vault_db):
    ret = vault_db.get_creds("testrole", cache=False)
    assert ret
    assert "username" in ret
    assert "password" in ret


@pytest.mark.usefixtures("role_static_setup")
def test_get_creds_static(vault_db, teststaticrole):
    ret = vault_db.get_creds("teststaticrole", static=True, cache=False)
    assert ret
    assert "username" in ret
    assert "password" in ret
    assert ret["username"] == teststaticrole["username"]


def test_get_creds_cached(vault_db, _cached_creds):
    ret_new = vault_db.get_creds("testrole", cache=True)
    assert ret_new
    assert "username" in ret_new
    assert "password" in ret_new
    assert ret_new["username"] == _cached_creds["username"]
    assert ret_new["password"] == _cached_creds["password"]


@pytest.mark.usefixtures("roles_setup")
def test_get_creds_cached__multiple(vault_db):
    ret = vault_db.get_creds("testrole", cache="one")
    assert ret
    assert "username" in ret
    assert "password" in ret
    ret_new = vault_db.get_creds("testrole", cache="two")
    assert ret_new
    assert "username" in ret_new
    assert "password" in ret_new
    assert ret_new["username"] != ret["username"]
    assert ret_new["password"] != ret["password"]
    assert vault_db.get_creds("testrole", cache="one") == ret
    assert vault_db.get_creds("testrole", cache="two") == ret_new


@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("roles_setup", [["testreissuerole"]], indirect=True)
@pytest.mark.parametrize(
    "_cached_creds", ({"role": "testreissuerole", "valid_for": 180},), indirect=True
)
def test_get_creds_cached_valid_for_reissue(vault_db, testreissuerole, _cached_creds):
    """
    Test that valid cached credentials that do not fulfill valid_for
    and cannot be renewed as requested are reissued
    """
    # 3 seconds because of leeway in lease validity check after renewals
    time.sleep(3)
    ret_new = vault_db.get_creds(
        "testreissuerole", cache=True, valid_for=testreissuerole["default_ttl"]
    )
    assert ret_new
    assert "username" in ret_new
    assert "password" in ret_new
    assert ret_new["username"] != _cached_creds["username"]
    assert ret_new["password"] != _cached_creds["password"]


@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("roles_setup", [["testreissuerole"]], indirect=True)
@pytest.mark.parametrize(
    "_cached_creds", ({"role": "testreissuerole", "valid_for": 180},), indirect=True
)
def test_get_creds_cached_with_cached_min_ttl(vault_db, _cached_creds):
    """
    Test that a cached ``min_ttl`` (``valid_for``) is respected at a minimum.
    """
    # 3 seconds because of leeway in lease validity check after renewals
    time.sleep(3)
    ret_new = vault_db.get_creds("testreissuerole", cache=True, valid_for=5)
    assert ret_new
    assert "username" in ret_new
    assert "password" in ret_new
    assert ret_new["username"] != _cached_creds["username"]
    assert ret_new["password"] != _cached_creds["password"]


@pytest.mark.parametrize(
    "_cached_creds,new",
    (
        ({"valid_for": 240}, {"valid_for": 360}),
        ({"revoke_delay": 240}, {"revoke_delay": 15}),
        ({"renew_increment": 240}, {"renew_increment": 15}),
        ({"meta": {"foo": "bar"}}, {"meta": {"bar": "baz"}}),
    ),
    indirect=("_cached_creds",),
)
@pytest.mark.parametrize("warn", (False, True))
def test_get_creds_cached_changed_lifecycle(vault_db, _cached_creds, new, warn, caplog):
    """
    Test that changed lifecycle attributes are warned about when
    _warn_about_attr_change is not set only.

    This is a precaution for the following situation:
    * A state manages the cached creds with wanted lifecycle attributes
    * During template rendering, the creds are requested with different
      lifecycle attributes.
    """
    with caplog.at_level(logging.WARNING):
        ret_new = vault_db.get_creds("testrole", cache=True, **new, _warn_about_attr_change=warn)
    assert ret_new
    assert "username" in ret_new
    assert "password" in ret_new
    assert ret_new["username"] == _cached_creds["username"]
    assert ret_new["password"] == _cached_creds["password"]
    assert ("changed lifecycle attributes" in caplog.text) is warn


@pytest.mark.usefixtures("_cached_creds")
def test_clear_cached(vault_db):
    assert vault_db.list_cached()
    assert vault_db.clear_cached() is True
    assert not vault_db.list_cached()


@pytest.mark.usefixtures("_cached_creds")
def test_list_cached(vault_db):
    ret = vault_db.list_cached()
    ckey = "db.database.dynamic.testrole.default"
    assert ret
    assert ckey in ret
    assert not ret[ckey]["expired"]
    assert ret[ckey]["expires_in"] > 3590
    assert "data" not in ret[ckey]
    now = datetime.now().astimezone()
    # this might fail if this test runs juuust before midnight
    assert ret[ckey]["creation_time"].startswith(now.strftime("%Y-%m-%d"))
    # I hope you have something better to do during New Year's Eve
    assert ret[ckey]["expire_time"].startswith(now.strftime("%Y-"))
    for val in ("creation_time", "expire_time"):
        assert ret[ckey][val].endswith(now.strftime(" %Z"))


@pytest.mark.usefixtures("_cached_creds")
def test_renew_cached(vault_db):
    curr = vault_db.list_cached()
    ckey = "db.database.dynamic.testrole.default"
    assert curr
    assert ckey in curr
    time.sleep(3)
    assert vault_db.renew_cached() is True
    new = vault_db.list_cached()
    assert new[ckey]["expire_time"] != curr[ckey]["expire_time"]


@pytest.mark.usefixtures("role_static_setup")
def test_rotate_static_role(vault_db):
    ret = vault_db.get_creds("teststaticrole", static=True, cache=False)
    assert ret
    old_pw = ret["password"]
    ret = vault_db.rotate_static_role("teststaticrole")
    assert ret
    ret = vault_db.get_creds("teststaticrole", static=True, cache=False)
    assert ret
    assert ret["password"] != old_pw
